# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from ..link_task import LinkTask

from .link_step_register_raw_dfs import LinkStepRegisterRawDfs
from .link_step_prep_dataframes import LinkStepPrepDataframes


class Preprocessing(LinkTask):
    def get_steps(self):
        return [LinkStepRegisterRawDfs(self), LinkStepPrepDataframes(self)]



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import pandas as pd
from hlink.linking.matching.link_step_score import LinkStepScore


def test_removal_of_duplicate_histid_b(
    spark, matching, matching_conf, scored_matches_test_data
):
    """Test all hh matching and training steps to ensure they work as a workflow"""
    path_pms = scored_matches_test_data
    matching.spark.read.csv(path_pms, header=True, inferSchema=True).write.mode(
        "overwrite"
    ).saveAsTable("scored_potential_matches")

    LinkStepScore(matching)._save_predicted_matches(
        matching_conf, "histid_a", "histid_b"
    )

    pred_matches = matching.spark.table("predicted_matches").toPandas()

    assert pred_matches.shape == (1, 3)
    assert (
        pred_matches.query("histid_a == 'A004' and histid_b == 'B003'")[
            "prediction"
        ].iloc[0]
        == 1
    )


def test_step_2_any_equals(spark, matching_household_conf, matching, preprocessing):

    matching_household_conf["column_mappings"].append(
        {
            "column_name": "namefrst_std",
            "alias": "namefrst_split",
            "transforms": [{"type": "split"}],
        }
    )

    matching_household_conf["column_mappings"].append(
        {
            "column_name": "namefrst_split",
            "alias": "namefrst_mid_init",
            "transforms": [
                {"type": "array_index", "value": 1},
                {"type": "substring", "values": [0, 1]},
            ],
        }
    )

    matching_household_conf["column_mappings"].append(
        {
            "column_name": "namefrst_split",
            "alias": "namefrst_unstd",
            "transforms": [{"type": "array_index", "value": 0}],
        }
    )

    matching_household_conf["comparison_features"] = [
        {
            "alias": "mid_init_match",
            "column_names": ["namefrst_mid_init", "namefrst_unstd"],
            "comparison_type": "any_equals",
        },
        {
            "alias": "namelast_jw",
            "column_name": "namelast_clean",
            "comparison_type": "jaro_winkler",
        },
    ]

    matching_household_conf["training"]["dependent_var"] = "match"
    matching_household_conf["training"]["independent_vars"] = [
        "neighbor_namelast_jw_rate",
        "neighbor_namelast_jw_rate_threshold",
        "namelast_jw",
    ]
    preprocessing.run_step(0)
    preprocessing.run_step(1)
    matching.run_step(0)
    matching.run_step(1)

    LinkStepScore(matching)._create_features(matching_household_conf)

    # Create pandas DFs of the step_2 potential matches table
    potential_matches_df = spark.table("potential_matches_prepped").toPandas()

    # Make assertions on the data
    assert len(potential_matches_df.id_a) == 37
    assert len(potential_matches_df.mid_init_match) == 37
    assert len(potential_matches_df.namefrst_mid_init_a == 37)
    assert (
        potential_matches_df.query(
            "id_a == '50b33ef6-259d-43af-8cdc-56a61f881169 ' and id_b == '50b33ef6-259d-43af-8cdc-56a61f881169 '"
        )["mid_init_match"].iloc[0]
        == 1
    )
    assert pd.isna(
        potential_matches_df.query(
            "id_a == '7fb55d25-2a7d-486d-9efa-27b9d7e60c24 ' and id_b == '7fb55d25-2a7d-486d-9efa-27b9d7e60c24 '"
        )["mid_init_match"].iloc[0]
    )
    assert (
        potential_matches_df.query(
            "id_a == '7fb55d25-2a7d-486d-9efa-27b9d7e60c24 ' and id_b == '7fb55d25-2a7d-486d-9efa-27b9d7e60c24 '"
        )["namefrst_mid_init_a"].iloc[0]
        is None
    )
    assert (
        potential_matches_df.query(
            "id_a == '7fb55d25-2a7d-486d-9efa-27b9d7e60c24 ' and id_b == '7fb55d25-2a7d-486d-9efa-27b9d7e60c24 '"
        )["namefrst_unstd_a"].iloc[0]
        == "phineas"
    )


def test_step_2_sum(spark, matching_household_conf, matching, preprocessing):

    matching_household_conf["feature_selections"] = [
        {
            "output_col": "namelast_popularity",
            "input_cols": ["sex", "bpl", "namelast_clean"],
            "transform": "popularity",
        }
    ]

    matching_household_conf["comparison_features"] = [
        {
            "alias": "namelast_popularity_sum",
            "column_name": "namelast_popularity",
            "comparison_type": "sum",
        },
        {
            "alias": "namelast_jw",
            "column_name": "namelast_clean",
            "comparison_type": "jaro_winkler",
        },
    ]

    matching_household_conf["training"]["dependent_var"] = "match"
    matching_household_conf["training"]["independent_vars"] = [
        "neighbor_namelast_jw_rate",
        "neighbor_namelast_jw_rate_threshold",
        "namelast_jw",
    ]
    preprocessing.run_step(0)
    preprocessing.run_step(1)
    matching.run_step(0)
    matching.run_step(1)

    LinkStepScore(matching)._create_features(matching_household_conf)

    # Create pandas DFs of the step_2 potential matches table
    potential_matches_df = spark.table("potential_matches_prepped").toPandas()
    assert len(potential_matches_df.namelast_popularity_a) == 37
    assert len(potential_matches_df.namelast_popularity_b) == 37
    assert len(potential_matches_df.namelast_popularity_sum) == 37
    assert (
        potential_matches_df.query(
            "id_a == 'ae7261c3-7d71-4ea1-997f-5d1a68c18777 ' and id_b == 'ad6442b5-42bc-4c2e-a517-5a951d989a92 '"
        )["namelast_popularity_a"].iloc[0]
        == 3
    )
    assert (
        potential_matches_df.query(
            "id_a == 'ae7261c3-7d71-4ea1-997f-5d1a68c18777 ' and id_b == 'ad6442b5-42bc-4c2e-a517-5a951d989a92 '"
        )["namelast_popularity_b"].iloc[0]
        == 2
    )
    assert (
        potential_matches_df.query(
            "id_a == 'ae7261c3-7d71-4ea1-997f-5d1a68c18777 ' and id_b == 'ad6442b5-42bc-4c2e-a517-5a951d989a92 '"
        )["namelast_popularity_sum"].iloc[0]
        == 5
    )


#
# TODO: fix hh compare rate java function
# def test_step_2_hh_compare_rate(spark, matching_household_conf, matching, preprocessing):
# matching_household_conf['feature_selections'] = [
#  {
#   "output_col": "namefrst_related_rows",
#  "input_cols": ["namefrst_std", "bpl", "sex"],
# "transform": "related_individual_rows",
# "family_id": "serialp",
# "relate_col": "relate",
# "top_code": 10,
# "bottom_code": 3
# }
# ]
#
#  matching_household_conf['comparison_features'] = [
#   {
#    "alias": "namelast_jw",
#   "column_name": "namelast_clean",
#  "comparison_type": "jaro_winkler"
# },
# {
# "alias": "related_match_rate",
# "column_name": "namefrst_related_rows",
# "comparison_type": "hh_compare_rate"
# }
# ]
#
# preprocessing.step_0_register_raw_dfs()
# preprocessing.step_1_prep_dataframe()
# matching.step_0_explode()
# matching.step_1_match()
#
# Create pandas DFs of the step_2 potential matches table
# potential_matches_df = spark.table("potential_matches").toPandas()
# assert (len(potential_matches_df.namelast_popularity_a) == 41)



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import itertools


def create_feature_tables(
    link_task, t_ctx_def, advanced_comp_features, hh_comp_features, id_col, table_name
):
    """Creates the table which contains all the comparison features from a
       table that has all the column and feature selections.

    Parameters
    ----------
    link_task: LinkTask
        the link task that is currently being ran
    t_ctx_def: dictionary
        the dictionary of values to pass to the sql templates
    advanced_comp_features: list
        a list of the "advanced" features which
        require potential match aggregation
    id_col: string
        the id column
    table_name: string
        the name of the output table

    Returns
    -------
    The output table with the features.
    """

    has_adv_comp_features = len(advanced_comp_features) > 0
    has_hh_comp_features = len(hh_comp_features) > 0
    gen_comp_feat = has_adv_comp_features or has_hh_comp_features
    config = link_task.link_run.config

    tmp_training_features = link_task.run_register_sql(
        name=f"tmp_{table_name}",
        template="potential_matches_base_features",
        t_ctx=t_ctx_def,
        persist=gen_comp_feat,
    )

    if has_adv_comp_features and not has_hh_comp_features:
        return link_task.run_register_sql(
            table_name,
            template="aggregate_features",
            t_ctx={
                "id": config["id_column"],
                "potential_matches": f"tmp_{table_name}",
                "advanced_comp_features": advanced_comp_features,
            },
            persist=True,
        )

    elif has_hh_comp_features and not has_adv_comp_features:
        return link_task.run_register_sql(
            table_name,
            template="hh_aggregate_features",
            t_ctx={
                "id": config["id_column"],
                "hh_col": config[f"{link_task.training_conf}"].get("hh_col", "serialp"),
                "potential_matches": f"tmp_{table_name}",
                "hh_comp_features": hh_comp_features,
            },
            persist=True,
        )

    elif has_adv_comp_features and has_hh_comp_features:
        af = link_task.run_register_sql(
            table_name,
            template="aggregate_features",
            t_ctx={
                "id": config["id_column"],
                "potential_matches": f"tmp_{table_name}",
                "advanced_comp_features": advanced_comp_features,
            },
        )
        return link_task.run_register_sql(
            af,
            template="hh_aggregate_features",
            t_ctx={
                "id": config["id_column"],
                "potential_matches": f"tmp_{table_name}",
                "hh_comp_features": hh_comp_features,
            },
            persist=True,
        )

    else:
        return link_task.run_register_python(
            table_name, lambda: tmp_training_features, persist=True
        )


def get_features(config, independent_features, pregen_features=[]):
    """Splits apart the comparison features into comp_features,
        advanced_features, and dist_features.

    Parameters
    ----------
    config: dictionary
        the base configuration dictionary
    independent_features: list
        a list of the comparison features to split apart
    pregen_features: dictionary
        a list of features that have been pregenerated and should be skipped

    Returns
    -------
    A 3-tuple of the standard comparison features,
    the advanced comparison features, and the distance features.
    """

    aggregate_features = [
        "hits",
        "hits2",
        "exact_mult",
        "exact_all_mult",
        "exact_all_mult2",
    ]

    hh_aggregate_features = ["jw_max_a", "jw_max_b"]

    all_comp_features = config["comparison_features"]
    advanced_comp_features = [
        f for f in all_comp_features if f["alias"] in aggregate_features
    ] + [f for f in independent_features if f in aggregate_features]
    hh_comp_features = [
        f for f in all_comp_features if f["alias"] in hh_aggregate_features
    ] + [f for f in independent_features if f in hh_aggregate_features]
    derived_comp_features = [
        f
        for f in all_comp_features
        if f["alias"] not in aggregate_features
        and f["alias"] not in pregen_features
        and f["alias"] not in hh_aggregate_features
    ]
    derived_aliases = [f["alias"] for f in derived_comp_features]

    dist_features = [
        c
        for c in all_comp_features
        if c["comparison_type"] == "geo_distance" and c["alias"] not in pregen_features
    ]

    if len({"exact_mult"} & set(advanced_comp_features)) > 0 and (
        ("exact" not in derived_aliases)
    ):
        raise KeyError(
            'In order to calculate "exact_mult", "exact" needs to be added to the list of comparison features in your configuration.'
        )

    if len(
        set(["exact_all_mult", "exact_all_mult2"]) & set(advanced_comp_features)
    ) > 0 and (("exact_all" not in derived_aliases)):
        raise KeyError(
            'In order to calculate "exact_all_mult", or "exact_all_mult2", "exact_all" needs to be added to the list of comparison features in your configuration.'
        )

    comp_features = ",\n ".join(
        [
            generate_comparison_feature(f, config["id_column"], include_as=True)
            for f in derived_comp_features
        ]
    )
    return comp_features, advanced_comp_features, hh_comp_features, dist_features


def generate_comparison_feature(feature, id_col, include_as=False):
    """Returns an SQL expression for a given feature.

    Parameters
    ----------
    feature: dictionary
        a comparison feature from the config
    id_col: string
        the id column
    include_as: boolean
        if true, then the expression will include "as {alias}",
        where `alias` is the alias of the given feature

    Returns
    -------
    A string containing the sql expression.
    """
    comp_type = feature["comparison_type"]

    if comp_type == "sql_condition":
        expr = feature["condition"]

    elif comp_type == "maximum_jaro_winkler":
        columns = feature["column_names"]
        comps = ", ".join(
            [
                f"jw(nvl(a.{col1}, ''), nvl(b.{col2}, ''))"
                for col1, col2 in itertools.product(columns, columns)
            ]
        )
        expr = f"GREATEST({comps})"

    elif comp_type == "jaro_winkler":
        col = feature["column_name"]
        expr = f"jw(nvl(a.{col}, ''), nvl(b.{col}, ''))"

    elif comp_type == "jaro_winkler_street":
        col = feature["column_name"]
        boundary_col = feature["boundary"]
        expr = f"IF(a.{boundary_col} = b.{boundary_col}, jw(nvl(a.{col},''), nvl(b.{col}, '')), 0)"

    elif comp_type == "max_jaro_winkler":
        col = feature["column_name"]
        expr = f"jw_max(a.{col}, b.{col})"

    elif comp_type == "equals":
        col = feature["column_name"]
        expr = f"a.{col} IS NOT DISTINCT FROM b.{col}"

    elif comp_type == "f1_match":
        fi = feature["first_init_col"]
        mi0 = feature["mid_init_cols"][0]
        mi1 = feature["mid_init_cols"][1]
        expr = (
            f"CASE WHEN ("
            f"(a.{fi} IS NOT DISTINCT FROM b.{fi}) OR "
            f"(a.{fi} IS NOT DISTINCT FROM b.{mi0}) OR "
            f"(a.{fi} IS NOT DISTINCT FROM b.{mi1})"
            f") THEN 1 ELSE 2 END"
        )

    elif comp_type == "f2_match":
        fi = feature["first_init_col"]
        mi0 = feature["mid_init_cols"][0]
        mi1 = feature["mid_init_cols"][1]
        expr = (
            f"CASE WHEN ((a.{mi0} == '') OR (a.{mi0} IS NULL)) THEN 0 WHEN ("
            f"(a.{mi0} IS NOT DISTINCT FROM b.{fi}) OR "
            f"((a.{mi1} IS NOT NULL) AND (a.{mi1} IS NOT DISTINCT FROM b.{fi})) OR "
            f"(a.{mi0} IS NOT DISTINCT FROM b.{mi0}) OR "
            f"(a.{mi0} IS NOT DISTINCT FROM b.{mi1}) OR "
            f"((a.{mi1} IS NOT NULL) AND (a.{mi1} IS NOT DISTINCT FROM b.{mi0})) OR "
            f"((a.{mi1} IS NOT NULL) AND (a.{mi1} IS NOT DISTINCT FROM b.{mi1}))"
            f") THEN 1 ELSE 2 END"
        )

    elif comp_type == "not_equals":
        col = feature["column_name"]
        expr = f"a.{col} IS DISTINCT FROM b.{col}"

    elif comp_type == "equals_as_int":
        col = feature["column_name"]
        expr = f"CAST(a.{col} = b.{col} as INT)"

    elif comp_type == "all_equals":
        cols = feature["column_names"]
        all_equals = " AND ".join([f"a.{col} = b.{col}" for col in cols])
        expr = f"{all_equals}"

    elif comp_type == "not_zero_and_not_equals":
        col = feature["column_name"]
        expr = f"a.{col} is not null and b.{col} is not null and a.{col} != 0 AND b.{col} != 0 and a.{col} IS DISTINCT FROM b.{col}"

    elif comp_type == "or":
        expr_a = generate_comparison_feature(feature["comp_a"], id_col)
        expr_b = generate_comparison_feature(feature["comp_b"], id_col)
        if "comp_d" in feature:
            expr_d = generate_comparison_feature(feature["comp_d"], id_col)
            expr_c = generate_comparison_feature(feature["comp_c"], id_col)
            expr = f"{expr_a}  OR {expr_b} OR {expr_c} OR {expr_d}"
        elif "comp_c" in feature:
            expr_c = generate_comparison_feature(feature["comp_c"], id_col)
            expr = f"{expr_a} OR {expr_b} OR {expr_c}"
        else:
            expr = f"{expr_a} OR {expr_b}"

    elif comp_type == "and":
        expr_a = generate_comparison_feature(feature["comp_a"], id_col)
        expr_b = generate_comparison_feature(feature["comp_b"], id_col)
        if "comp_d" in feature:
            expr_d = generate_comparison_feature(feature["comp_d"], id_col)
            expr_c = generate_comparison_feature(feature["comp_c"], id_col)
            expr = f"{expr_a}  AND {expr_b} AND {expr_c} AND {expr_d}"
        elif "comp_c" in feature:
            expr_c = generate_comparison_feature(feature["comp_c"], id_col)
            expr = f"{expr_a} AND {expr_b} AND {expr_c}"
        else:
            expr = f"{expr_a} AND {expr_b}"

    elif comp_type == "times":
        expr_a = generate_comparison_feature(feature["comp_a"], id_col)
        expr_b = generate_comparison_feature(feature["comp_b"], id_col)
        expr = f"CAST({expr_a} as float) * CAST({expr_b} as float)"

    elif comp_type == "caution_comp_3":
        expr_a = generate_comparison_feature(feature["comp_a"], id_col)
        expr_b = generate_comparison_feature(feature["comp_b"], id_col)
        expr_c = generate_comparison_feature(feature["comp_c"], id_col)
        expr = f"({expr_a}  OR {expr_b}) AND {expr_c}"

    elif comp_type == "caution_comp_4":
        expr_a = generate_comparison_feature(feature["comp_a"], id_col)
        expr_b = generate_comparison_feature(feature["comp_b"], id_col)
        expr_c = generate_comparison_feature(feature["comp_c"], id_col)
        expr_d = generate_comparison_feature(feature["comp_d"], id_col)
        expr = f"({expr_a}  OR {expr_b} OR {expr_c}) AND {expr_d}"

    elif comp_type == "caution_comp_3_012":
        expr_a = generate_comparison_feature(feature["comp_a"], id_col)
        expr_b = generate_comparison_feature(feature["comp_b"], id_col)
        expr_c = generate_comparison_feature(feature["comp_c"], id_col)
        expr = (
            f"CASE WHEN CAST({expr_c} as string) == 'false' THEN 2 "
            f"WHEN ({expr_a} OR {expr_b}) AND {expr_c} THEN 1 "
            f"ELSE 0 END"
        )

    elif comp_type == "caution_comp_4_012":
        expr_a = generate_comparison_feature(feature["comp_a"], id_col)
        expr_b = generate_comparison_feature(feature["comp_b"], id_col)
        expr_c = generate_comparison_feature(feature["comp_c"], id_col)
        expr_d = generate_comparison_feature(feature["comp_d"], id_col)
        expr = (
            f"CASE WHEN CAST({expr_d} as string) == 'false' THEN 2 "
            f"WHEN (({expr_a}) OR ({expr_b}) OR ({expr_c})) AND ({expr_d}) THEN 1 "
            f"ELSE 0 END"
        )

    elif comp_type == "any_equals":
        col1, col2 = feature["column_names"]
        expr = f"""
        (
          ( a.{col1} = b.{col1} OR a.{col1} = b.{col2} )
          AND
          nvl(a.{col1}, '') != ''
        )
        OR
        (
          ( a.{col2} = b.{col1})
          AND
          nvl(a.{col2}, '') != ''
        )
        """

    elif comp_type == "either_are_1":
        col = feature["column_name"]
        expr = f"(a.{col} = 1 OR b.{col} = 1)"

    elif comp_type == "either_are_0":
        col = feature["column_name"]
        expr = f"(a.{col} = 0 OR b.{col} = 0)"

    elif comp_type == "second_gen_imm":
        col = feature["column_name"]
        expr = f"(a.{col} = 2 OR a.{col} = 3 OR a.{col} = 4)"

    elif comp_type == "rel_jaro_winkler":
        col = feature["column_name"]
        if "jw_threshold" in feature:
            jw_threshold = feature["jw_threshold"]
        else:
            jw_threshold = 0.8
            print(
                f"WARNING: No jw_threshold defined; Setting jw_threshold for rel_jaro_winkler comparison feature for {col} to {jw_threshold}"
            )
        if feature.get("age_threshold", False):
            age_threshold = feature["age_threshold"]
        else:
            age_threshold = 5
            print(
                f"WARNING: No age_threshold defined; Setting age_threshold for rel_jaro_winkler comparison feature for {col} to {age_threshold}"
            )
        histid = feature.get("histid_col", "histid")
        name = feature.get("name_col", "namefrst_std")
        byr = feature.get("birthyr_col", "birthyr")
        sex = feature.get("sex_col", "sex")
        expr = f"rel_jw(a.{col}, b.{col}, string({jw_threshold}), string({age_threshold}), map('name','{name}','byr','{byr}','sex','{sex}'))"

    elif comp_type == "extra_children":
        col = feature["column_name"]
        if "jw_threshold" in feature:
            jw_threshold = feature["jw_threshold"]
        else:
            jw_threshold = 0.8
            print(
                f"WARNING: No jw_threshold defined; Setting jw_threshold for rel_jaro_winkler comparison feature for {col} to {jw_threshold}"
            )
        if feature.get("age_threshold", False):
            age_threshold = feature["age_threshold"]
        else:
            age_threshold = 5
            print(
                f"WARNING: No age_threshold defined; Setting age_threshold for rel_jaro_winkler comparison feature for {col} to {age_threshold}"
            )
        year_b = feature.get("year_b", "year_b")
        relate = feature.get("relate_col", "relate")
        histid = id_col if id_col is not None else "histid"
        name = feature.get("name_col", "namefrst_std")
        byr = feature.get("birthyr_col", "birthyr")
        sex = feature.get("sex_col", "sex")
        expr = f"extra_children(a.{col}, b.{col}, string({year_b}), a.{relate}, b.{relate}, string({jw_threshold}), string({age_threshold}), map('histid', '{histid}', 'name','{name}','byr','{byr}','sex','{sex}'))"

    elif comp_type == "jaro_winkler_rate":
        col = feature["column_name"]
        if "jw_threshold" in feature:
            jw_threshold = feature["jw_threshold"]
        else:
            jw_threshold = 0.8
            print(
                f"WARNING: No jw_threshold defined; Setting jw_threshold for jaro_winkler_rate comparison feature for {col} to {jw_threshold}"
            )
        expr = f"jw_rate(a.{col}, b.{col}, string({jw_threshold}))"

    elif comp_type == "sum":
        col = feature["column_name"]
        expr = f"a.{col} + b.{col}"

    elif comp_type == "hh_compare_rate":
        col = feature["column_name"]
        expr = f"hh_compare_rate(a.{col}, b.{col})"

    elif comp_type == "length_b":
        col = feature["column_name"]
        expr = f"size(b.{col})"

    elif comp_type == "abs_diff":
        col = feature["column_name"]
        ne = feature.get("not_equals", False)
        if ne:
            expr = f"case when cast(b.{col} as INT) != {ne} and cast(a.{col} as INT) != {ne} then abs(CAST(b.{col} as INT) - CAST(a.{col} as INT)) else -1 end"
        else:
            expr = f"abs(CAST(b.{col} as INT) - CAST(a.{col} as INT))"

    elif comp_type == "b_minus_a":
        col = feature["column_name"]
        ne = feature.get("not_equals", False)
        if ne:
            expr = f"case when cast(b.{col} as INT) != {ne} and cast(a.{col} as INT) != {ne} then CAST(b.{col} as INT) - CAST(a.{col} as INT) else -1 end"
        else:
            expr = f"CAST(b.{col} as INT) - CAST(a.{col} as INT)"

    elif comp_type == "has_matching_element":
        col = feature["column_name"]
        expr = f"has_matching_element(a.{col}, b.{col})"
    elif comp_type == "geo_distance":
        distance_col = feature["distance_col"]
        dt = feature["table_name"]
        st = feature.get("secondary_table_name", False)
        if st:
            st_distance_col = feature["secondary_distance_col"]
            expr = f"IF({dt}.{distance_col} IS NOT NULL, {dt}.{distance_col}, {st}.{st_distance_col})"
        else:
            expr = f"{dt}.{distance_col}"

    elif comp_type == "fetch_a":
        col = feature["column_name"]
        expr = f"a.{col}"

    elif comp_type == "fetch_b":
        col = feature["column_name"]
        expr = f"b.{col}"

    elif comp_type == "fetch_td":
        col = feature["column_name"]
        expr = f"pm.{col}"

    elif comp_type == "new_marr":
        col = feature["column_name"]
        if "upper_threshold" not in feature:
            feature["upper_threshold"] = 10
            print(
                f"WARNING: No upper_threshold defined; Setting upper_threshold for new_marr comparison feature for {col} to 10"
            )
        expr = f"CAST(b.{col} as INT)"

    elif comp_type == "existing_marr":
        col = feature["column_name"]
        if "lower_threshold" not in feature:
            feature["lower_threshold"] = 10
            print(
                f"WARNING: No lower_threshold defined; Setting lower_threshold for existing_marr comparison feature for {col} to 10"
            )
        expr = f"CAST(b.{col} as INT)"

    elif comp_type == "parent_step_change":
        col = feature["column_name"]
        expr = f"(CAST(a.{col} as INT) > 0) IS DISTINCT FROM (CAST(b.{col} as INT) > 0)"

    elif comp_type == "present_both_years":
        col = feature["column_name"]
        expr = f"a.{col} IS NOT NULL AND a.{col} > 0 AND b.{col} IS NOT NULL AND b.{col} > 0"

    elif comp_type == "neither_are_null":
        col = feature["column_name"]
        expr = f"a.{col} IS NOT NULL AND b.{col} IS NOT NULL and a.{col} != '' and b.{col} != ''"

    elif comp_type == "present_and_matching_categorical":
        col = feature["column_name"]
        expr = f"IF(a.{col} IS NOT NULL AND b.{col} IS NOT NULL AND CAST(a.{col} as STRING) != '' and CAST(b.{col} as STRING) != '', IF(a.{col} IS DISTINCT FROM b.{col}, 1, 0), 2)"

    elif comp_type == "present_and_equal_categorical_in_universe":
        col = feature["column_name"]
        niu = feature["NIU"]
        expr = f"IF(a.{col} IS NOT NULL AND b.{col} IS NOT NULL AND a.{col} != {niu} AND b.{col} != {niu} AND CAST(a.{col} as string) != '' and CAST(b.{col} as string) != '', IF(a.{col} IS DISTINCT FROM b.{col}, 0, 1), 0)"

    elif comp_type == "present_and_not_equal":
        col = feature["column_name"]
        expr = f"IF(a.{col} IS NOT NULL AND b.{col} IS NOT NULL AND cast(a.{col} as string) != '' and cast(b.{col} as string) != '' and a.{col} > 0, IF(a.{col} IS DISTINCT FROM b.{col}, TRUE, FALSE), FALSE)"

    else:
        raise ValueError(f"No comparison type: {feature['comparison_type']}")

    if feature.get("power", False):
        exponent = feature["power"]
        expr = f"POWER(CAST({expr} as INT), {exponent})"

    if feature.get("threshold", False):
        threshold = feature["threshold"]
        expr = f"{expr} IS NOT NULL and {expr} >= {threshold}"
    elif feature.get("lower_threshold", False):
        lower_threshold = feature["lower_threshold"]
        expr = f"{expr} IS NOT NULL and {expr} >= {lower_threshold}"
    elif feature.get("upper_threshold", False):
        upper_threshold = feature["upper_threshold"]
        expr = f"{expr} IS NOT NULL and {expr} <= {upper_threshold}"
    elif feature.get("gt_threshold", False):
        gt_threshold = feature["gt_threshold"]
        expr = f"{expr} IS NOT NULL and {expr} > {gt_threshold}"
    elif feature.get("btwn_threshold", False):
        bt0 = feature["btwn_threshold"][0]
        bt1 = feature["btwn_threshold"][1]
        expr = f"{expr} IS NOT NULL and {expr} >= {bt0} and {expr} <= {bt1}"

    if feature.get("look_at_addl_var", False):
        addl_var = feature["addl_var"]
        check_val_expr = feature["check_val_expr"]
        else_val = feature["else_val"]
        datasource = feature["datasource"]
        expr = f"CASE WHEN {datasource}.{addl_var} {check_val_expr} then {expr} else {else_val} END"

    if include_as:
        full_expr = f"({expr})" + f" as {feature['alias']}"
    else:
        full_expr = expr

    return full_expr



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import itertools
import math
import re
import numpy as np
import pandas as pd
from sklearn.metrics import precision_recall_curve, auc
from pyspark.sql.functions import count, mean

import hlink.linking.core.threshold as threshold_core
import hlink.linking.core.classifier as classifier_core

from hlink.linking.link_step import LinkStep


class LinkStepTrainTestModels(LinkStep):
    def __init__(self, task):
        super().__init__(
            task,
            "train test models",
            input_table_names=[
                f"{task.table_prefix}training_features",
                f"{task.table_prefix}training_vectorized",
            ],
            output_table_names=[
                f"{task.table_prefix}training_results",
                f"{task.table_prefix}repeat_fps",
                f"{task.table_prefix}repeat_fns",
                f"{task.table_prefix}repeat_tps",
                f"{task.table_prefix}repeat_tns",
            ],
        )

    def _run(self):
        training_conf = str(self.task.training_conf)
        table_prefix = self.task.table_prefix
        config = self.task.link_run.config

        self.task.spark.sql("set spark.sql.shuffle.partitions=1")

        dep_var = config[training_conf]["dependent_var"]
        id_a = config["id_column"] + "_a"
        id_b = config["id_column"] + "_b"
        desc_df = _create_desc_df()
        columns_to_keep = [id_a, id_b, "features_vector", dep_var]
        prepped_data = (
            self.task.spark.table(f"{table_prefix}training_vectorized")
            .select(columns_to_keep)
            .cache()
        )

        otd_data = self._create_otd_data(id_a, id_b)

        n_training_iterations = config[training_conf].get("n_training_iterations", 10)
        seed = config[training_conf].get("seed", 2133)

        splits = self._get_splits(prepped_data, id_a, n_training_iterations, seed)

        model_parameters = self._get_model_parameters(config)
        for run in model_parameters:
            params = run.copy()
            model_type = params.pop("type")

            alpha_threshold = params.pop(
                "threshold", config[training_conf].get("threshold", 0.8)
            )
            if (
                config[training_conf].get("decision", False)
                == "drop_duplicate_with_threshold_ratio"
            ):
                threshold_ratio = params.pop(
                    "threshold_ratio",
                    threshold_core.get_threshold_ratio(config[training_conf], params),
                )
            else:
                threshold_ratio = False

            threshold_matrix = _calc_threshold_matrix(alpha_threshold, threshold_ratio)
            results_dfs = {}
            for i in range(len(threshold_matrix)):
                results_dfs[i] = _create_results_df()

            first = True
            for training_data, test_data in splits:
                training_data.cache()
                test_data.cache()

                classifier, post_transformer = classifier_core.choose_classifier(
                    model_type, params, dep_var
                )

                model = classifier.fit(training_data)

                predictions_tmp = _get_probability_and_select_pred_columns(
                    test_data, model, post_transformer, id_a, id_b, dep_var
                ).cache()
                predict_train_tmp = _get_probability_and_select_pred_columns(
                    training_data, model, post_transformer, id_a, id_b, dep_var
                ).cache()

                test_pred = predictions_tmp.toPandas()
                precision, recall, thresholds_raw = precision_recall_curve(
                    test_pred[f"{dep_var}"],
                    test_pred["probability"].round(2),
                    pos_label=1,
                )

                thresholds_plus_1 = np.append(thresholds_raw, [np.nan])
                param_text = np.full(precision.shape, model_type + "_" + str(params))

                pr_auc = auc(recall, precision)
                print(f"Area under PR curve: {pr_auc}")

                if first:
                    prc = pd.DataFrame(
                        {
                            "params": param_text,
                            "precision": precision,
                            "recall": recall,
                            "threshold_gt_eq": thresholds_plus_1,
                        }
                    )
                    self.task.spark.createDataFrame(prc).write.mode(
                        "overwrite"
                    ).saveAsTable(
                        f"{self.task.table_prefix}precision_recall_curve_"
                        + re.sub("[^A-Za-z0-9]", "_", model_type + str(params))
                    )

                    first = False

                i = 0
                for at, tr in threshold_matrix:
                    predictions = threshold_core.predict_using_thresholds(
                        predictions_tmp,
                        at,
                        tr,
                        config[training_conf],
                        config["id_column"],
                    )
                    predict_train = threshold_core.predict_using_thresholds(
                        predict_train_tmp,
                        at,
                        tr,
                        config[training_conf],
                        config["id_column"],
                    )

                    results_dfs[i] = self._capture_results(
                        predictions,
                        predict_train,
                        dep_var,
                        model,
                        results_dfs[i],
                        otd_data,
                        at,
                        tr,
                        pr_auc,
                    )
                    i += 1

                training_data.unpersist()
                test_data.unpersist()

            for i in range(len(threshold_matrix)):
                desc_df = _append_results(desc_df, results_dfs[i], model_type, params)

        _print_desc_df(desc_df)
        desc_df = _load_desc_df_params(desc_df)
        self._save_training_results(desc_df, self.task.spark)
        self._save_otd_data(otd_data, self.task.spark)
        self.task.spark.sql("set spark.sql.shuffle.partitions=200")

    def _get_splits(self, prepped_data, id_a, n_training_iterations, seed):
        if self.task.link_run.config[f"{self.task.training_conf}"].get(
            "split_by_id_a", False
        ):
            split_ids = [
                prepped_data.select(id_a)
                .distinct()
                .randomSplit([0.5, 0.5], seed=seed + i)
                for i in range(n_training_iterations)
            ]

            splits = []
            for ids_a, ids_b in split_ids:
                split_a = prepped_data.join(ids_a, on=id_a, how="inner")
                split_b = prepped_data.join(ids_b, on=id_a, how="inner")
                splits.append([split_a, split_b])

        else:
            splits = [
                prepped_data.randomSplit([0.5, 0.5], seed=seed + i)
                for i in range(n_training_iterations)
            ]

        return splits

    def _custom_param_grid_builder(self, conf):
        print("Building param grid for models")
        given_parameters = conf[f"{self.task.training_conf}"]["model_parameters"]
        new_params = []
        for run in given_parameters:
            params = run.copy()
            model_type = params.pop("type")

            # dropping thresholds to prep for scikitlearn model exploration refactor
            threshold = params.pop("threshold", False)
            threshold_ratio = params.pop("threshold_ratio", False)

            keys = params.keys()
            values = params.values()

            params_exploded = []
            for prod in itertools.product(*values):
                params_exploded.append(dict(zip(keys, prod)))

            for subdict in params_exploded:
                subdict["type"] = model_type
                if threshold:
                    subdict["threshold"] = threshold
                if threshold_ratio:
                    subdict["threshold_ratio"] = threshold_ratio

            new_params.extend(params_exploded)
        return new_params

    def _capture_results(
        self,
        predictions,
        predict_train,
        dep_var,
        model,
        results_df,
        otd_data,
        at,
        tr,
        pr_auc,
    ):
        table_prefix = self.task.table_prefix

        print("Evaluating model performance...")
        # write to sql tables for testing
        predictions.createOrReplaceTempView(f"{table_prefix}predictions")
        predict_train.createOrReplaceTempView(f"{table_prefix}predict_train")

        (
            test_TP_count,
            test_FP_count,
            test_FN_count,
            test_TN_count,
        ) = _get_confusion_matrix(predictions, dep_var, otd_data)
        test_precision, test_recall, test_mcc = _get_aggregate_metrics(
            test_TP_count, test_FP_count, test_FN_count, test_TN_count
        )

        (
            train_TP_count,
            train_FP_count,
            train_FN_count,
            train_TN_count,
        ) = _get_confusion_matrix(predict_train, dep_var, otd_data)
        train_precision, train_recall, train_mcc = _get_aggregate_metrics(
            train_TP_count, train_FP_count, train_FN_count, train_TN_count
        )

        new_results = pd.DataFrame(
            {
                "precision_test": [test_precision],
                "recall_test": [test_recall],
                "precision_train": [train_precision],
                "recall_train": [train_recall],
                "pr_auc": [pr_auc],
                "test_mcc": [test_mcc],
                "train_mcc": [train_mcc],
                "model_id": [model],
                "alpha_threshold": [at],
                "threshold_ratio": [tr],
            },
        )
        return pd.concat([results_df, new_results], ignore_index=True)

    def _get_model_parameters(self, conf):
        training_conf = str(self.task.training_conf)

        model_parameters = conf[training_conf]["model_parameters"]
        if "param_grid" in conf[training_conf] and conf[training_conf]["param_grid"]:
            model_parameters = self._custom_param_grid_builder(conf)
        elif model_parameters == []:
            raise ValueError(
                "No model parameters found. In 'training' config, either supply 'model_parameters' or 'param_grid'."
            )
        return model_parameters

    def _save_training_results(self, desc_df, spark):
        table_prefix = self.task.table_prefix

        if desc_df.empty:
            print("Training results dataframe is empty.")
        else:
            desc_df.dropna(axis=1, how="all", inplace=True)
            spark.createDataFrame(desc_df, samplingRatio=1).write.mode(
                "overwrite"
            ).saveAsTable(f"{table_prefix}training_results")
            print(
                f"Training results saved to Spark table '{table_prefix}training_results'."
            )

    def _prepare_otd_table(self, spark, df, id_a, id_b):
        spark_df = spark.createDataFrame(df)
        counted = (
            spark_df.groupby(id_a, id_b)
            .agg(
                count("*").alias("count"),
                mean("probability").alias("mean_probability"),
            )
            .filter("count > 1")
            .orderBy(["count", id_a, id_b])
        )
        return counted

    def _save_otd_data(self, otd_data, spark):
        table_prefix = self.task.table_prefix

        if otd_data is None:
            return
        id_a = otd_data["id_a"]
        id_b = otd_data["id_b"]

        if not otd_data["FP_data"].empty:
            table_name = f"{table_prefix}repeat_fps"
            counted_FPs = self._prepare_otd_table(
                spark, otd_data["FP_data"], id_a, id_b
            )
            counted_FPs.write.mode("overwrite").saveAsTable(table_name)
            print(
                f"A table of false positives of length {counted_FPs.count()} was saved as '{table_name}' for analysis."
            )
        else:
            print("There were no false positives recorded.")

        if not otd_data["FN_data"].empty:
            table_name = f"{table_prefix}repeat_fns"
            counted_FNs = self._prepare_otd_table(
                spark, otd_data["FN_data"], id_a, id_b
            )
            counted_FNs.write.mode("overwrite").saveAsTable(table_name)
            print(
                f"A table of false negatives of length {counted_FNs.count()} was saved as '{table_name}' for analysis."
            )
        else:
            print("There were no false negatives recorded.")

        if not otd_data["TP_data"].empty:
            table_name = f"{table_prefix}repeat_tps"
            counted_TPs = self._prepare_otd_table(
                spark, otd_data["TP_data"], id_a, id_b
            )
            counted_TPs.write.mode("overwrite").saveAsTable(table_name)
            print(
                f"A table of true positives of length {counted_TPs.count()} was saved as '{table_name}' for analysis."
            )
        else:
            print("There were no true positives recorded.")

        if not otd_data["TN_data"].empty:
            table_name = f"{table_prefix}repeat_tns"
            counted_TNs = self._prepare_otd_table(
                spark, otd_data["TN_data"], id_a, id_b
            )
            counted_TNs.write.mode("overwrite").saveAsTable(table_name)
            print(
                f"A table of true negatives of length {counted_TNs.count()} was saved as '{table_name}' for analysis."
            )
        else:
            print("There were no true negatives recorded.")

    def _create_otd_data(self, id_a, id_b):
        """Output Suspicous Data (OTD): used to check config to see if you should find sketchy training data that the models routinely mis-classify"""
        training_conf = str(self.task.training_conf)
        config = self.task.link_run.config

        if (
            "output_suspicious_TD" in config[training_conf]
            and config[training_conf]["output_suspicious_TD"]
        ):
            return {
                "FP_data": pd.DataFrame(),
                "FN_data": pd.DataFrame(),
                "TP_data": pd.DataFrame(),
                "TN_data": pd.DataFrame(),
                "id_a": id_a,
                "id_b": id_b,
            }
        else:
            return None


def _calc_mcc(TP, TN, FP, FN):
    if (math.sqrt((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN))) != 0:
        mcc = ((TP * TN) - (FP * FN)) / (
            math.sqrt((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN))
        )
    else:
        mcc = 0
    return mcc


def _calc_threshold_matrix(alpha_threshold, threshold_ratio):
    if alpha_threshold and type(alpha_threshold) != list:
        alpha_threshold = [alpha_threshold]

    if threshold_ratio and type(threshold_ratio) != list:
        threshold_ratio = [threshold_ratio]

    if threshold_ratio:
        threshold_matrix = [[a, b] for a in alpha_threshold for b in threshold_ratio]
    else:
        threshold_matrix = [[a, np.nan] for a in alpha_threshold]

    return threshold_matrix


def _get_probability_and_select_pred_columns(
    pred_df, model, post_transformer, id_a, id_b, dep_var
):
    all_prediction_cols = set(
        [
            f"{id_a}",
            f"{id_b}",
            dep_var,
            "probability",
            "probability_array",
            "prediction",
        ]
    )
    transformed_df = model.transform(pred_df)
    post_transform_df = post_transformer.transform(transformed_df)
    required_col_df = post_transform_df.select(
        list(all_prediction_cols & set(post_transform_df.columns))
    )
    return required_col_df


def _get_confusion_matrix(predictions, dep_var, otd_data):
    TP = predictions.filter((predictions[dep_var] == 1) & (predictions.prediction == 1))
    TP_count = TP.count()

    FP = predictions.filter((predictions[dep_var] == 0) & (predictions.prediction == 1))
    FP_count = FP.count()

    FN = predictions.filter((predictions[dep_var] == 1) & (predictions.prediction == 0))
    FN_count = FN.count()

    TN = predictions.filter((predictions[dep_var] == 0) & (predictions.prediction == 0))
    TN_count = TN.count()

    if otd_data:
        id_a = otd_data["id_a"]
        id_b = otd_data["id_b"]

        new_FP_data = FP.select(
            id_a, id_b, dep_var, "prediction", "probability"
        ).toPandas()
        otd_data["FP_data"] = pd.concat([otd_data["FP_data"], new_FP_data])

        new_FN_data = FN.select(
            id_a, id_b, dep_var, "prediction", "probability"
        ).toPandas()
        otd_data["FN_data"] = pd.concat([otd_data["FN_data"], new_FN_data])

        new_TP_data = TP.select(
            id_a, id_b, dep_var, "prediction", "probability"
        ).toPandas()
        otd_data["TP_data"] = pd.concat([otd_data["TP_data"], new_TP_data])

        new_TN_data = TN.select(
            id_a, id_b, dep_var, "prediction", "probability"
        ).toPandas()
        otd_data["TN_data"] = pd.concat([otd_data["TN_data"], new_TN_data])

    return TP_count, FP_count, FN_count, TN_count


def _get_aggregate_metrics(TP_count, FP_count, FN_count, TN_count):
    if (TP_count + FP_count) == 0:
        precision = np.nan
    else:
        precision = TP_count / (TP_count + FP_count)
    if (TP_count + FN_count) == 0:
        recall = np.nan
    else:
        recall = TP_count / (TP_count + FN_count)
    mcc = _calc_mcc(TP_count, TN_count, FP_count, FN_count)
    return precision, recall, mcc


def _create_results_df():
    return pd.DataFrame(
        columns=[
            "precision_test",
            "recall_test",
            "precision_train",
            "recall_train",
            "pr_auc",
            "test_mcc",
            "train_mcc",
            "model_id",
            "alpha_threshold",
            "threshold_ratio",
        ]
    )


def _append_results(desc_df, results_df, model_type, params):
    # run.pop("type")
    print(results_df)

    new_desc = pd.DataFrame(
        {
            "model": [model_type],
            "parameters": [params],
            "alpha_threshold": [results_df["alpha_threshold"][0]],
            "threshold_ratio": [results_df["threshold_ratio"][0]],
            "precision_test_mean": [results_df["precision_test"].mean()],
            "precision_test_sd": [results_df["precision_test"].std()],
            "recall_test_mean": [results_df["recall_test"].mean()],
            "recall_test_sd": [results_df["recall_test"].std()],
            "pr_auc_mean": [results_df["pr_auc"].mean()],
            "pr_auc_sd": [results_df["pr_auc"].std()],
            "mcc_test_mean": [results_df["test_mcc"].mean()],
            "mcc_test_sd": [results_df["test_mcc"].std()],
            "precision_train_mean": [results_df["precision_train"].mean()],
            "precision_train_sd": [results_df["precision_train"].std()],
            "recall_train_mean": [results_df["recall_train"].mean()],
            "recall_train_sd": [results_df["recall_train"].std()],
            "mcc_train_mean": [results_df["train_mcc"].mean()],
            "mcc_train_sd": [results_df["train_mcc"].std()],
        },
    )

    desc_df = pd.concat([desc_df, new_desc], ignore_index=True)
    _print_desc_df(desc_df)
    return desc_df


def _print_desc_df(desc_df):
    pd.set_option("display.max_colwidth", None)
    print(
        desc_df.drop(
            [
                "recall_test_sd",
                "recall_train_sd",
                "precision_test_sd",
                "precision_train_sd",
            ],
            axis=1,
        ).iloc[-1]
    )
    print("\n")


def _load_desc_df_params(desc_df):
    params = [
        "maxDepth",
        "numTrees",
        "featureSubsetStrategy",
        "subsample",
        "minInstancesPerNode",
        "maxBins",
        "class_weight",
        "C",
        "kernel",
        "threshold",
        "maxIter",
    ]

    load_params = lambda j, param: j.get(param, np.nan)
    for param in params:
        desc_df[param] = desc_df["parameters"].apply(load_params, args=(param,))
    desc_df["class_weight"] = desc_df["class_weight"].apply(
        lambda x: str(x) if pd.notnull(x) else x
    )
    desc_df["parameters"] = desc_df["parameters"].apply(
        lambda t: str(t) if pd.notnull(t) else t
    )
    return desc_df


def _create_desc_df():
    return pd.DataFrame(
        columns=[
            "model",
            "parameters",
            "alpha_threshold",
            "threshold_ratio",
            "precision_test_mean",
            "precision_test_sd",
            "recall_test_mean",
            "recall_test_sd",
            "mcc_test_mean",
            "mcc_test_sd",
            "precision_train_mean",
            "precision_train_sd",
            "recall_train_mean",
            "recall_train_sd",
            "pr_auc_mean",
            "pr_auc_sd",
            "mcc_train_mean",
            "mcc_train_sd",
        ]
    )



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink


def get_blocking(conf):
    if "blocking" in conf:
        return conf["blocking"]
    else:
        print(
            "DEPRECATION WARNING: The config value 'blocking_steps' has been renamed to 'blocking' and is now just a single array of objects."
        )
        return conf["blocking_steps"][0]



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from pyspark.sql.functions import col, lit
import hlink.linking.core.transforms as transforms_core


def select_column_mapping(column_mapping, df_selected, is_a, column_selects):
    name = column_mapping["column_name"]
    if "override_column_a" in column_mapping and is_a:
        override_name = column_mapping["override_column_a"]
        column_select = col(override_name)
        if "override_transforms" in column_mapping:
            for transform in column_mapping["override_transforms"]:
                column_select = transforms_core.apply_transform(
                    column_select, transform, is_a
                )
    elif "override_column_b" in column_mapping and not is_a:
        override_name = column_mapping["override_column_b"]
        column_select = col(override_name)
        if "override_transforms" in column_mapping:
            for transform in column_mapping["override_transforms"]:
                column_select = transforms_core.apply_transform(
                    column_select, transform, is_a
                )
    elif "set_value_column_a" in column_mapping and is_a:
        value_to_set = column_mapping["set_value_column_a"]
        column_select = lit(value_to_set)
    elif "set_value_column_b" in column_mapping and not is_a:
        value_to_set = column_mapping["set_value_column_b"]
        column_select = lit(value_to_set)
    elif "transforms" in column_mapping:
        column_select = col(name)
        for transform in column_mapping["transforms"]:
            column_select = transforms_core.apply_transform(
                column_select, transform, is_a
            )
    else:
        column_select = col(name)

    alias = column_mapping["alias"] if "alias" in column_mapping else name

    column_selects.append(alias)
    return df_selected.withColumn(alias, column_select), column_selects



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from pyspark.ml.util import JavaMLReadable, JavaMLWritable
from pyspark.ml.param.shared import HasInputCols, HasOutputCol
from pyspark import keyword_only
from pyspark.ml.wrapper import JavaTransformer


class InteractionTransformer(
    JavaTransformer, HasInputCols, HasOutputCol, JavaMLReadable, JavaMLWritable
):
    """
    from https://github.com/apache/spark/commit/5bf5d9d854db53541956dedb03e2de8eecf65b81:
    Implements the feature interaction transform. This transformer takes in Double and Vector type
    columns and outputs a flattened vector of their feature interactions. To handle interaction,
    we first one-hot encode any nominal features. Then, a vector of the feature cross-products is
    produced.
    For example, given the input feature values `Double(2)` and `Vector(3, 4)`, the output would be
    `Vector(6, 8)` if all input features were numeric. If the first feature was instead nominal
    with four categories, the output would then be `Vector(0, 0, 0, 0, 3, 4, 0, 0)`.
    df = spark.createDataFrame([(0.0, 1.0), (2.0, 3.0)], ["a", "b"])
    interaction = Interaction(inputCols=["a", "b"], outputCol="ab")
    interaction.transform(df).show()
    +---+---+-----+
    |  a|  b|   ab|
    +---+---+-----+
    |0.0|1.0|[0.0]|
    |2.0|3.0|[6.0]|
    +---+---+-----+
    ...
    interactionPath = temp_path + "/interaction"
    interaction.save(interactionPath)
    loadedInteraction = Interaction.load(interactionPath)
    loadedInteraction.transform(df).head().ab == interaction.transform(df).head().ab
    True
    .. versionadded:: 3.0.0
    """

    @keyword_only
    def __init__(self, inputCols=None, outputCol=None):
        """
        __init__(self, inputCols=None, outputCol=None):
        """
        super(InteractionTransformer, self).__init__()
        self._java_obj = self._new_java_obj(
            "org.apache.spark.ml.feature.Interaction", self.uid
        )
        self._setDefault()
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @keyword_only
    def setParams(self, inputCols=None, outputCol=None):
        """
        setParams(self, inputCols=None, outputCol=None)
         for this Interaction.
        """
        kwargs = self._input_kwargs
        return self._set(**kwargs)



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import sys
from pathlib import Path

from hlink.spark.session import SparkConnection


class SparkFactory:
    """This class allows convenient creation of a spark session.

    It defines defaults for many settings that can be overwritten with the
    `set_*` functions. Each `set_*` function returns the SparkFactory for easy
    chaining of settings.

    Note that some settings values are paths. These paths should be absolute.
    This applies to derby_dir, warehouse_dir, and tmp_dir.
    """

    def __init__(self):
        spark_dir = Path("spark").resolve()
        self.derby_dir = spark_dir / "derby"
        self.warehouse_dir = spark_dir / "warehouse"
        self.tmp_dir = spark_dir / "tmp"
        self.python = sys.executable
        self.db_name = "linking"
        self.is_local = True
        self.num_cores = 4
        self.executor_memory = "10G"
        self.executor_cores = 16

    def set_derby_dir(self, derby_dir):
        self.derby_dir = derby_dir
        return self

    def set_warehouse_dir(self, warehouse_dir):
        self.warehouse_dir = warehouse_dir
        return self

    def set_tmp_dir(self, tmp_dir):
        self.tmp_dir = tmp_dir
        return self

    def set_python(self, python):
        """Set the python executable.

        Useful when you want to guarantee that remote machines are running the
        same version of python.
        """
        self.python = python
        return self

    def set_db_name(self, db_name):
        self.db_name = db_name
        return self

    def set_local(self):
        """Make a local spark connection."""
        self.is_local = True
        return self

    def set_num_cores(self, num_cores):
        self.num_cores = num_cores
        return self

    def set_executor_memory(self, executor_memory):
        self.executor_memory = executor_memory
        return self

    def set_executor_cores(self, executor_cores):
        self.executor_cores = executor_cores
        return self

    def create(self):
        spark_conn = SparkConnection(
            str(self.derby_dir),
            str(self.warehouse_dir),
            str(self.tmp_dir),
            self.python,
            self.db_name,
        )
        return spark_conn.local(self.num_cores, self.executor_memory)



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import pytest
from pyspark.ml import Pipeline
import hlink.linking.core.pipeline as pipeline_core


@pytest.mark.quickcheck
def test_all_steps(
    spark,
    training_conf,
    training,
    state_dist_path,
    training_data_path,
    potential_matches_path,
    spark_test_tmp_dir_path,
    matching,
    training_validation_path,
):
    """Test running the chosen model on potential matches dataset"""
    training_conf["comparison_features"] = [
        {
            "alias": "regionf",
            "column_name": "region",
            "comparison_type": "fetch_a",
            "categorical": True,
        },
        {
            "alias": "namelast_jw",
            "column_name": "namelast",
            "comparison_type": "jaro_winkler",
        },
        {
            "alias": "state_distance",
            "column_name": "bpl",
            "key_count": 1,
            "comparison_type": "geo_distance",
            "loc_a": "statecode1",
            "loc_b": "statecode2",
            "distance_col": "dist",
            "table_name": "state_distances_lookup",
            "distances_file": state_dist_path,
        },
        {
            "alias": "exact",
            "column_names": ["namelast", "namelast"],
            "comparison_type": "all_equals",
        },
        {
            "alias": "exact_all",
            "column_names": ["namelast", "bpl", "sex"],
            "comparison_type": "all_equals",
        },
    ]

    training_conf["training"]["dataset"] = training_data_path
    training_conf["training"]["dependent_var"] = "match"
    training_conf["training"]["independent_vars"] = [
        "namelast_jw",
        "regionf",
        "state_distance",
        "exact_mult",
        "exact_all_mult",
        "hits",
        "namelast_jw_buckets",
    ]
    training_conf["pipeline_features"] = [
        {
            "input_column": "namelast_jw",
            "output_column": "namelast_jw_buckets",
            "transformer_type": "bucketizer",
            "categorical": True,
            "splits": [0, 0.25, 0.5, 0.75, 0.99, 1],
        }
    ]
    training_conf["training"]["chosen_model"] = {
        "type": "random_forest",
        "maxDepth": 6,
        "numTrees": 100,
        "featureSubsetStrategy": "sqrt",
    }

    # training_conf["training"]["use_potential_matches_features"] = True
    training_conf["training"]["score_with_model"] = True
    training_conf["spark_tmp_dir"] = spark_test_tmp_dir_path

    training.link_run.trained_models["trained_model"] = None

    training.run_step(0)

    training.run_step(1)
    tf = spark.table("training_features").toPandas()
    assert tf.query("id_a == 20 and id_b == 30")["exact"].iloc[0]
    assert not tf.query("id_a == 20 and id_b == 50")["exact"].iloc[0]
    assert not tf.query("id_a == 20 and id_b == 30")["exact_mult"].iloc[0]
    assert not tf.query("id_a == 20 and id_b == 10")["exact_mult"].iloc[0]

    training.run_step(2)

    # m = PipelineModel.load(spark_test_tmp_dir_path + "/chosen_model")
    p = training.link_run.trained_models["pre_pipeline"]
    m = training.link_run.trained_models["trained_model"]
    transformed_df = m.transform(
        p.transform(spark.table("training_features"))
    ).toPandas()
    row = transformed_df.query("id_a == 10 and id_b == 50").iloc[0]
    assert row.prediction == 0
    assert row.state_distance_imp.round(0) == 1909


def test_step_2_bucketizer(spark, main, conf):
    """Test a bucketized feature using spark pipeline function"""
    data = [
        (0.0, 0, 0),
        (3.0, 1, 1),
        (5.0, 2, 0),
        (6.0, 3, 1),
        (9.0, 4, 0),
        (10.0, 5, 1),
        (11.0, 6, 0),
        (23.0, 7, 1),
    ]
    dataFrame = spark.createDataFrame(data, ["immyear_abs_diff", "test_id", "match"])
    dataFrame.createOrReplaceTempView("training_features")

    conf["pipeline_features"] = [
        {
            "input_column": "immyear_abs_diff",
            "output_column": "immyear_caution",
            "transformer_type": "bucketizer",
            "categorical": True,
            "splits": [0, 6, 11, 999],
        }
    ]
    conf["training"] = {
        "dependent_var": "match",
        "independent_vars": ["immyear_abs_diff", "immyear_caution"],
    }
    conf["comparison_features"] = []

    ind_vars = conf["training"]["independent_vars"]
    tf = spark.table("training_features")
    pipeline_stages = pipeline_core.generate_pipeline_stages(
        conf, ind_vars, tf, "training"
    )
    prep_pipeline = Pipeline(stages=pipeline_stages)
    prep_model = prep_pipeline.fit(tf)
    prepped_data = prep_model.transform(tf)
    prepped_data = prepped_data.toPandas()

    assert prepped_data.shape == (8, 7)
    assert list(prepped_data.query("test_id == 0")["features_vector"].iloc[0]) == [
        0,
        1,
        0,
        0,
        0,
    ]
    assert list(prepped_data.query("test_id == 1")["features_vector"].iloc[0]) == [
        3,
        1,
        0,
        0,
        0,
    ]
    assert list(prepped_data.query("test_id == 3")["features_vector"].iloc[0]) == [
        6,
        0,
        1,
        0,
        0,
    ]
    assert list(prepped_data.query("test_id == 6")["features_vector"].iloc[0]) == [
        11,
        0,
        0,
        1,
        0,
    ]

    main.do_drop_all("")


def test_step_2_interaction(spark, main, conf):
    """Test interacting two and three features using spark pipeline function"""
    data = [
        (0.0, 0.0, 0.0),
        (3.0, 1.0, 1.0),
        (5.0, 2.0, 0.0),
        (6.0, 3.0, 1.0),
        (9.0, 4.0, 0.0),
        (10.0, 5.0, 1.0),
        (11.0, 6.0, 0.0),
        (23.0, 7.0, 1.0),
    ]
    dataFrame = spark.createDataFrame(data, ["var0", "var1", "var2"])
    dataFrame.createOrReplaceTempView("training_features")

    conf["pipeline_features"] = [
        {
            "input_columns": ["var0", "var1"],
            "output_column": "interacted_vars01",
            "transformer_type": "interaction",
        },
        {
            "input_columns": ["var0", "var1", "var2"],
            "output_column": "interacted_vars012",
            "transformer_type": "interaction",
        },
    ]

    conf["training"] = {
        "dependent_var": "var2",
        "independent_vars": ["interacted_vars01", "interacted_vars012"],
    }
    conf["comparison_features"] = []

    ind_vars = conf["training"]["independent_vars"]
    tf = spark.table("training_features")
    pipeline_stages = pipeline_core.generate_pipeline_stages(
        conf, ind_vars, tf, "training"
    )
    prep_pipeline = Pipeline(stages=pipeline_stages)
    prep_model = prep_pipeline.fit(tf)
    prepped_data = prep_model.transform(tf)
    prepped_data = prepped_data.toPandas()

    assert prepped_data.shape == (8, 8)
    assert prepped_data.query("var1 == 0")["interacted_vars01"].iloc[0][0] == 0
    assert prepped_data.query("var1 == 2")["interacted_vars01"].iloc[0][0] == 10
    assert prepped_data.query("var1 == 2")["interacted_vars012"].iloc[0][0] == 0
    assert prepped_data.query("var1 == 3")["interacted_vars01"].iloc[0][0] == 18
    assert prepped_data.query("var1 == 3")["interacted_vars012"].iloc[0][0] == 18

    main.do_drop_all("")



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from hlink.linking.link_step import LinkStep


class LinkStepIngestFile(LinkStep):
    def __init__(self, task):
        super().__init__(
            task,
            "ingest file",
            input_table_names=[],
            output_table_names=[f"{task.table_prefix}training_data"],
        )

    def _run(self):
        self.task.run_register_python(
            f"{self.task.table_prefix}training_data",
            lambda: self.task.spark.read.csv(
                self.task.link_run.config[f"{self.task.training_conf}"]["dataset"],
                header=True,
                inferSchema=True,
            ),
            persist=True,
        )



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import os
import pandas as pd
import pytest
from pyspark.sql.types import StructType, StructField, LongType
from hlink.errors import DataError


@pytest.mark.quickcheck
def test_step_0(preprocessing, spark, preprocessing_conf):
    """Test preprocessing step 0 to ensure that temporary raw_df_unpartitioned_(a/b) tables are created (exact copies of datasources from config). Also test that the presistent raw_df_(a/b) tables are created. Should be same as raw datasources with filters applied"""

    # Run the preprocessing step (this will use the test data)
    preprocessing.run_step(0)

    # Create pandas DFs of the step_0 preprocessed test data
    up_pdf_a = spark.table("raw_df_unpartitioned_a").toPandas()
    up_pdf_b = spark.table("raw_df_unpartitioned_b").toPandas()

    pdf_a = spark.table("raw_df_a").toPandas()
    pdf_b = spark.table("raw_df_b").toPandas()

    # Make assertions on the data
    assert pdf_a.query("id == 10")["serialp"].iloc[0] == "A"
    assert pdf_a.query("id == 20")["serialp"].iloc[0] == "B"

    assert pdf_b.query("id == 10")["serialp"].iloc[0] == "C"
    assert pdf_b.query("id == 30")["serialp"].iloc[0] == "D"

    assert up_pdf_a.query("id == 10")["serialp"].iloc[0] == "A"
    assert up_pdf_a.query("id == 30")["serialp"].iloc[0] == "B"

    assert up_pdf_b.query("id == 30")["serialp"].iloc[0] == "D"
    assert up_pdf_b.query("id == 50")["serialp"].iloc[0] == "E"


def test_step_0_datasource_parquet_file(
    preprocessing, spark, preprocessing_conf, input_data_dir_path
):
    """Test preprocessing step 0 to ensure that temporary raw_df_unpartitioned_(a/b) tables are created (exact copies of datasources from config). Also test that the presistent raw_df_(a/b) tables are created. Should be same as raw datasources with filters applied"""

    preprocessing_conf["datasource_a"] = {
        "parquet_file": os.path.join(
            input_data_dir_path, "test_parquet_data_a.parquet"
        ),
        "alias": "parquet_file_test_a",
    }

    # Run the preprocessing step (this will use the test data)
    preprocessing.run_step(0)

    # Create pandas DFs of the step_0 preprocessed test data
    up_pdf_a = spark.table("raw_df_unpartitioned_a").toPandas()

    pdf_a = spark.table("raw_df_a").toPandas()

    # Make assertions on the data
    assert pdf_a.query("id == 10")["bpl"].iloc[0] == 120
    assert pdf_a.query("id == 20")["bpl"].iloc[0] == 240

    assert up_pdf_a.query("id == 30")["bpl"].iloc[0] == 360
    assert up_pdf_a.query("id == 20")["bpl"].iloc[0] == 240


def test_step_0_datasource_file_parquet(
    preprocessing, spark, preprocessing_conf, input_data_dir_path
):
    """Test preprocessing step 0 to ensure that temporary raw_df_unpartitioned_(a/b) tables are created (exact copies of datasources from config). Also test that the presistent raw_df_(a/b) tables are created. Should be same as raw datasources with filters applied"""

    preprocessing_conf["datasource_a"] = {
        "file": os.path.join(input_data_dir_path, "test_parquet_data_b.parquet"),
        "alias": "parquet_file_test_b",
    }

    # Run the preprocessing step (this will use the test data)
    preprocessing.run_step(0)

    # Create pandas DFs of the step_0 preprocessed test data
    up_pdf_a = spark.table("raw_df_unpartitioned_a").toPandas()

    pdf_a = spark.table("raw_df_a").toPandas()

    # Make assertions on the data
    assert pdf_a.query("id == 10")["bpl"].iloc[0] == 460
    assert pdf_a.query("id == 30")["bpl"].iloc[0] == 540

    assert up_pdf_a.query("id == 10")["bpl"].iloc[0] == 460
    assert up_pdf_a.query("id == 50")["bpl"].iloc[0] == 710


def test_step_0_datasource_file_csv(
    preprocessing, spark, preprocessing_conf, input_data_dir_path
):
    """Test preprocessing step 0 to ensure that temporary raw_df_unpartitioned_(a/b) tables are created (exact copies of datasources from config). Also test that the presistent raw_df_(a/b) tables are created. Should be same as raw datasources with filters applied"""

    preprocessing_conf["datasource_a"] = {
        "file": os.path.join(input_data_dir_path, "test_csv_data_a.csv"),
        "alias": "csv_file_test_a",
    }
    preprocessing_conf["datasource_b"] = {
        "file": os.path.join(input_data_dir_path, "test_csv_data_b.csv"),
        "alias": "csv_file_test_b",
    }

    # Run the preprocessing step (this will use the test data)
    preprocessing.run_step(0)

    # Create pandas DFs of the step_0 preprocessed test data
    up_pdf_a = spark.table("raw_df_unpartitioned_a").toPandas()
    up_pdf_b = spark.table("raw_df_unpartitioned_b").toPandas()

    pdf_a = spark.table("raw_df_a").toPandas()
    pdf_b = spark.table("raw_df_b").toPandas()

    # Make assertions on the data
    assert pdf_a.query("id == 10")["bpl"].iloc[0] == 120
    assert pdf_a.query("id == 30")["bpl"].iloc[0] == 360

    assert pdf_b.query("id == 30")["bpl"].iloc[0] == 540
    assert pdf_b.query("id == 50")["bpl"].iloc[0] == 710

    assert up_pdf_a.query("id == 20")["bpl"].iloc[0] == 240
    assert up_pdf_a.query("id == 10")["bpl"].iloc[0] == 120

    assert up_pdf_b.query("id == 10")["bpl"].iloc[0] == 460
    assert up_pdf_b.query("id == 30")["bpl"].iloc[0] == 540


def test_step_0_filters_training_data(preprocessing, spark, preprocessing_conf):
    """Test filter run in preprocessing step 0 which selects any person in a household which has a person who is also represented in the test data"""

    # create some training data we can use for testing
    td_schema = StructType(
        [StructField("id_a", LongType(), True), StructField("id_b", LongType(), True)]
    )
    data_training = [{"id_a": 10, "id_b": 30}, {"id_a": 20, "id_b": 40}]
    td = spark.createDataFrame(data_training, schema=td_schema)

    # create the training_data table
    preprocessing.run_register_python(
        name="training_data", func=lambda: td, persist=True
    )

    # add a filter to the config to filter only for households in the training data set
    preprocessing_conf["filter"] = [{"training_data_subset": True}]
    assert preprocessing_conf["filter"] != []

    # run the preprocessing step which includes filtering as a function
    preprocessing.run_step(0)

    # Create pandas DFs of the step_0 preprocessed test data
    pdf_a = spark.table("raw_df_a").toPandas()
    pdf_b = spark.table("raw_df_b").toPandas()

    # Make assertions on the data
    assert len(pdf_a.id) == 3
    assert len(pdf_b.id) == 1
    assert pdf_a.query("id == 20")["namelast"].iloc[0] == "Mc Last"
    assert pd.isnull(pdf_b.query("id == 30")["namelast"].iloc[0])


def test_step_0_filters_expression(preprocessing, spark, preprocessing_conf):
    """Test a filter run in preprocessing step 0 which selects rows from the raw data according to an expression"""

    # overwrite the config filter value to include only an expression type filter
    preprocessing_conf["filter"] = [
        {"expression": "namelast is not null and namelast != ''"}
    ]
    assert preprocessing_conf["filter"] != []

    # run the preprocessing step which includes filtering as a function
    preprocessing.run_step(0)

    # create pandas DFs of the step_0 preprocessed test data
    pdf_a = spark.table("raw_df_a").toPandas()
    pdf_b = spark.table("raw_df_b").toPandas()

    # make assertions on the data
    assert len(pdf_a.id) == 2
    assert len(pdf_b.id) == 1
    assert not pdf_a.namelast.isnull().values.any()
    assert not pdf_b.namelast.isnull().values.any()


def test_step_0_filters_expression_and_household(
    preprocessing, spark, preprocessing_conf
):
    """Test a filter run in preprocessing step 0 which selects rows from the raw data according to an expression AND includes any other entries with the same household ID (serialp)"""

    # overwrite the config filter value to include an expression filter which includes a household:true argument.
    # Note: in a household filter, the variables for household ID in each input file must be specified in the filter.  (This is the "serial_a": "serialp" bit below.)

    preprocessing_conf["filter"] = [
        {
            "expression": "namelast == 'Mc Last' or namelast == 'Name'",
            "household": True,
            "serial_a": "serialp",
            "serial_b": "serialp",
        }
    ]

    # run the preprocessing step which includes filtering as a function
    preprocessing.run_step(0)

    # create pandas DFs of the step_0 preprocessed test data
    pdf_a = spark.table("raw_df_a").toPandas()
    pdf_b = spark.table("raw_df_b").toPandas()

    # make assertions on the data
    assert len(pdf_a.id) == 2
    assert len(pdf_b.id) == 1
    assert not pdf_a.namelast.isnull().values.any()
    assert not pdf_b.namelast.isnull().values.any()


def test_step_0_filters_datasource(preprocessing, spark, preprocessing_conf):
    """Test a filter run in preprocessing step 0 which selects rows from the raw data according
    to an expression AND only applies the expression to a specified datasource (a or b)"""

    # overwrite the config filter value to include an expression filter which includes a datasource argument.
    preprocessing_conf["filter"] = [
        {"expression": "id == 30", "datasource": "a"},
        {"expression": "id == 10", "datasource": "b"},
    ]
    assert preprocessing_conf["filter"] != []

    # run the preprocessing step which includes filtering as a function
    preprocessing.run_step(0)

    # create pandas DFs of the step_0 preprocessed test data
    pdf_a = spark.table("raw_df_a").toPandas()
    pdf_b = spark.table("raw_df_b").toPandas()

    # make assertions on the data
    assert len(pdf_a.id) == 1
    assert len(pdf_b.id) == 1
    assert pdf_a.id[0] == 30
    assert pdf_b.id[0] == 10


def test_step_0_check_for_all_spaces_unrestricted_data(
    preprocessing, spark, preprocessing_conf_all_space_columns, capsys
):
    """Tests the check in preprocessing that looks for all-space columns, as found in unrestricted data files."""
    with pytest.raises(DataError, match=r"\: namelast, street\."):
        preprocessing.run_step(0)


def test_step_1_transform_attach_variable(
    preprocessing, spark, preprocessing_conf, region_code_path
):
    """Test the transform "attach_variable" -- used to add a feature column from CSV data"""
    preprocessing_conf["column_mappings"] = [{"column_name": "bpl"}]
    preprocessing_conf["feature_selections"] = [
        {
            "input_column": "bpl",
            "output_column": "region",
            "transform": "attach_variable",
            "region_dict": region_code_path,
            "col_to_join_on": "bpl",
            "col_to_add": "region",
            "null_filler": 99,
            "col_type": "int",
        }
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert len(pdf_a.region) == 3
    assert len(pdf_b.region) == 3
    assert pdf_a.query("id == 10")["region"].iloc[0] == 6
    assert pdf_a.query("id == 30")["region"].iloc[0] == 99
    assert pdf_b.query("id == 10")["region"].iloc[0] == 8
    assert pdf_b.query("id == 50")["region"].iloc[0] == 99


def test_step_1_transform_hash(preprocessing, spark, preprocessing_conf):
    """Test the transform "attach_variable" -- used to add a feature column from CSV data"""
    preprocessing_conf["column_mappings"].append(
        {
            "column_name": "namefrst",
            "alias": "namefrst_clean",
            "transforms": [
                {"type": "lowercase_strip"},
                {"type": "rationalize_name_words"},
                {"type": "remove_qmark_hyphen"},
                {"type": "replace_apostrophe"},
                {
                    "type": "remove_suffixes",
                    "values": ["jr", "sr", "ii", "iii", "iv", "v", "vi", "vii", "viii"],
                },
                {"type": "remove_alternate_names"},
                {"type": "condense_strip_whitespace"},
            ],
        }
    )

    preprocessing_conf["feature_selections"] = [
        {
            "input_column": "namefrst_clean",
            "output_column": "namefrst_bigrams",
            "transform": "bigrams",
            "no_first_pad": True,
        },
        {
            "input_column": "namefrst_bigrams",
            "output_column": "namefrst_bigrams_hash",
            "transform": "hash",
            "number": 5,
        },
    ]

    preprocessing_conf["filter"] = [
        {"expression": "namefrst is not null and namefrst != ''"}
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert len(pdf_a.namefrst_bigrams_hash_count) == 3
    assert len(pdf_a.namefrst_bigrams_hash) == 3

    assert len(pdf_b.namefrst_bigrams_hash_count) == 2
    assert len(pdf_b.namefrst_bigrams_hash) == 2

    assert len(pdf_a.query("id == 10")["namefrst_bigrams_hash_count"].iloc[0]) == 17
    assert len(pdf_a.query("id == 10")["namefrst_bigrams_hash"].iloc[0]) == 5

    assert len(pdf_b.query("id == 10")["namefrst_bigrams_hash_count"].iloc[0]) == 6
    assert len(pdf_b.query("id == 10")["namefrst_bigrams_hash"].iloc[0]) == 5


def test_step_1_transform_override(
    preprocessing, spark, preprocessing_conf, region_code_path
):
    """Test a column mapping with transform and OVERRIDE column for a specified datasource -- used to generate a feature from a column from one datasource and use a pre-existing column from the other datasource"""
    preprocessing_conf["column_mappings"] = [
        {
            "column_name": "namefrst",
            "alias": "namefrst_mid_init",
            "override_column_b": "namemiddle",
            "transforms": [
                {"type": "lowercase_strip"},
                {"type": "rationalize_name_words"},
                {"type": "remove_qmark_hyphen"},
                {"type": "replace_apostrophe"},
                {
                    "type": "remove_suffixes",
                    "values": ["jr", "sr", "ii", "iii", "iv", "v", "vi", "vii", "viii"],
                },
                {"type": "remove_prefixes", "values": ["mr"]},
                {"type": "remove_alternate_names"},
                {"type": "condense_strip_whitespace"},
                {"type": "split"},
                {"type": "array_index", "value": 1},
            ],
            "override_transforms": [
                {"type": "lowercase_strip"},
                {"type": "rationalize_name_words"},
                {"type": "remove_qmark_hyphen"},
                {"type": "replace_apostrophe"},
                {
                    "type": "remove_suffixes",
                    "values": ["jr", "sr", "ii", "iii", "iv", "v", "vi", "vii", "viii"],
                },
                {"type": "remove_alternate_names"},
                {"type": "condense_strip_whitespace"},
            ],
        }
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert pdf_a.query("id == 10")["namefrst_mid_init"].iloc[0] == "m"
    assert pdf_a.query("id == 20")["namefrst_mid_init"].iloc[0] == "marc"
    assert pd.isnull(pdf_a.query("id == 30")["namefrst_mid_init"].iloc[0])
    assert pdf_b.query("id == 10")["namefrst_mid_init"].iloc[0] == "m"
    assert pdf_b.query("id == 50")["namefrst_mid_init"].iloc[0] == "marc"
    assert pd.isnull(pdf_b.query("id == 30")["namefrst_mid_init"].iloc[0])


def test_step_1_transforms_namefrst_soundex(
    preprocessing, spark, preprocessing_conf, region_code_path
):
    """Test a column mapping with string-based transforms on firstname removing middle name"""
    preprocessing_conf["column_mappings"] = [
        {
            "column_name": "namefrst",
            "alias": "namefrst_std",
            "transforms": [
                {"type": "lowercase_strip"},
                {"type": "rationalize_name_words"},
                {"type": "remove_qmark_hyphen"},
                {"type": "replace_apostrophe"},
                {
                    "type": "remove_suffixes",
                    "values": ["jr", "sr", "ii", "iii", "iv", "v", "vi", "vii", "viii"],
                },
                {"type": "remove_prefixes", "values": ["mr"]},
                {"type": "remove_alternate_names"},
                {"type": "remove_one_letter_names"},
                {"type": "condense_strip_whitespace"},
                {"type": "split"},
                {"type": "array_index", "value": 0},
            ],
        }
    ]

    preprocessing_conf["feature_selections"] = [
        {
            "input_column": "namefrst_std",
            "output_column": "namefrst_soundex",
            "transform": "soundex",
        }
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert pdf_a.query("id == 10")["namefrst_std"].iloc[0] == "john"
    assert pdf_a.query("id == 20")["namefrst_std"].iloc[0] == "marc"
    assert pdf_a.query("id == 30")["namefrst_std"].iloc[0] == "jon"
    assert pdf_a.query("namefrst_std == 'john'")["namefrst_soundex"].iloc[0] == "J500"
    assert pdf_a.query("namefrst_std == 'marc'")["namefrst_soundex"].iloc[0] == "M620"

    assert pdf_b.query("id == 10")["namefrst_std"].iloc[0] == "john"
    assert pd.isnull(pdf_b.query("id == 30")["namefrst_std"].iloc[0])
    assert pdf_b.query("id == 50")["namefrst_std"].iloc[0] == "jean"
    assert pdf_b.query("namefrst_std == 'jean'")["namefrst_soundex"].iloc[0] == "J500"


def test_step_1_transforms_prefix_suffix(
    preprocessing, spark, preprocessing_conf, region_code_path
):
    """Test a column mapping with different string transforms"""
    preprocessing_conf["column_mappings"] = [
        {
            "column_name": "namefrst",
            "alias": "namefrst_std",
            "transforms": [
                {"type": "lowercase_strip"},
                {"type": "rationalize_name_words"},
                {"type": "remove_qmark_hyphen"},
                {"type": "replace_apostrophe"},
                {
                    "type": "remove_suffixes",
                    "values": ["jr", "sr", "ii", "iii", "iv", "v", "vi", "vii", "viii"],
                },
                {"type": "remove_prefixes", "values": ["mr"]},
                {"type": "remove_alternate_names"},
                {"type": "condense_strip_whitespace"},
            ],
        }
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert pdf_a.query("id == 10")["namefrst_std"].iloc[0] == "john m"
    assert pdf_a.query("id == 20")["namefrst_std"].iloc[0] == "j marc ell"
    assert pdf_a.query("id == 30")["namefrst_std"].iloc[0] == "jon"

    assert pdf_b.query("id == 10")["namefrst_std"].iloc[0] == "john"
    assert pd.isnull(pdf_b.query("id == 30")["namefrst_std"].iloc[0])
    assert pdf_b.query("id == 50")["namefrst_std"].iloc[0] == "jean"


def test_step_1_transform_adv_str(preprocessing, spark, preprocessing_conf):
    """Test a column mapping with remaining transforms"""
    preprocessing_conf["column_mappings"] = [
        {
            "column_name": "namelast",
            "alias": "namelast_std",
            "transforms": [
                {"type": "lowercase_strip"},
                {
                    "type": "condense_prefixes",
                    "values": ["mc", "mac", "o", "de", "van", "di"],
                },
            ],
        },
        {"column_name": "sex"},
        {
            "column_name": "sex",
            "alias": "sex_int",
            "transforms": [{"type": "cast_as_int"}],
        },
        {
            "column_name": "namelast_std",
            "alias": "namelast_std_len",
            "transforms": [{"type": "length"}],
        },
        {
            "alias": "namelast_init",
            "column_name": "namelast_std",
            "transforms": [{"type": "substring", "values": [0, 1]}],
        },
        {
            "column_name": "sex_int",
            "alias": "sex_mapped",
            "transforms": [{"type": "mapping", "mappings": {"1": "M", "2": "F"}}],
        },
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()

    assert pdf_a.query("id == 20")["namelast_std"].iloc[0] == "mclast"
    assert pdf_a.query("id == 10")["sex_int"].iloc[0] == 1
    assert pdf_a.query("id == 20")["namelast_std_len"].iloc[0] == 6
    assert pdf_a.query("id == 20")["namelast_init"].iloc[0] == "m"
    assert pdf_a.query("id == 30")["namelast_init"].iloc[0] == "l"
    assert pdf_a.query("id == 10")["sex_mapped"].iloc[0] == "M"
    assert pdf_a.query("id == 20")["sex_mapped"].iloc[0] == "F"


def test_step_1_transform_neighbor_agg(
    preprocessing, spark, preprocessing_conf_household_data
):
    """Test neighbor_aggregate transform on data containing households"""
    preprocessing_conf_household_data["feature_selections"] = [
        {
            "output_column": "namelast_neighbors",
            "input_column": "namelast_clean",
            "transform": "neighbor_aggregate",
            "neighborhood_column": "enumdist",
            "sort_column": "serial",
            "range": 5,
        }
    ]

    preprocessing.run_step(0)

    rda = spark.table("raw_df_a")
    rda = (
        rda.withColumn("enumdist_tmp", rda["enumdist"].cast("bigint"))
        .drop("enumdist")
        .withColumnRenamed("enumdist_tmp", "enumdist")
    )
    rda = (
        rda.withColumn("serial_tmp", rda["serial"].cast("bigint"))
        .drop("serial")
        .withColumnRenamed("serial_tmp", "serial")
    )
    rda = (
        rda.withColumn("pernum_tmp", rda["pernum"].cast("bigint"))
        .drop("pernum")
        .withColumnRenamed("pernum_tmp", "pernum")
    )
    rda.write.mode("overwrite").saveAsTable("raw_df_a_tmp")
    spark.sql("drop table raw_df_a")
    spark.sql("alter table raw_df_a_tmp rename to raw_df_a")

    rdb = spark.table("raw_df_b")
    rdb = (
        rdb.withColumn("enumdist_tmp", rdb["enumdist"].cast("bigint"))
        .drop("enumdist")
        .withColumnRenamed("enumdist_tmp", "enumdist")
    )
    rdb = (
        rdb.withColumn("serial_tmp", rdb["serial"].cast("bigint"))
        .drop("serial")
        .withColumnRenamed("serial_tmp", "serial")
    )
    rdb = (
        rdb.withColumn("pernum_tmp", rdb["pernum"].cast("bigint"))
        .drop("pernum")
        .withColumnRenamed("pernum_tmp", "pernum")
    )
    rdb.write.mode("overwrite").saveAsTable("raw_df_b_tmp")
    spark.sql("drop table raw_df_b")
    spark.sql("alter table raw_df_b_tmp rename to raw_df_b")

    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    wilson_a_nbrs = sorted(
        pdf_a.query("namelast_clean == 'wilson'")["namelast_neighbors"].iloc[0]
    )
    wilson_b_nbrs = sorted(
        pdf_b.query("namelast_clean == 'wilson'")["namelast_neighbors"].iloc[0]
    )
    lord_a_nbrs = sorted(
        pdf_a.query("namelast_clean == 'lord'")["namelast_neighbors"].iloc[0]
    )

    assert lord_a_nbrs == ["allen", "dekay", "foster", "graham", "taylor", "thorpe"]

    assert wilson_a_nbrs == [
        "bierhahn",
        "chambers",
        "cleveland",
        "collins",
        "flemming",
        "graham",
        "harvey",
        "mclean",
        "seward",
        "shields",
    ]
    assert wilson_b_nbrs == [
        "bierhahn",
        "cleveland",
        "collins",
        "dekay",
        "flemming",
        "graham",
        "harvey",
        "mclean",
        "seward",
        "shields",
    ]


def test_step_1_transform_attach_family_col(
    preprocessing, spark, preprocessing_conf_household_data
):
    """Test attach_family_col transform on data containing households"""
    preprocessing_conf_household_data["feature_selections"] = [
        {
            "output_col": "spouse_namefrst",
            "transform": "attach_family_col",
            "other_col": "namefrst_clean",
            "person_pointer": "sploc",
            "family_id": "serial",
            "person_id": "pernum",
        },
        {
            "output_col": "spouse_bpl",
            "transform": "attach_family_col",
            "other_col": "bpl",
            "person_pointer": "sploc",
            "family_id": "serial",
            "person_id": "pernum",
        },
        {
            "output_col": "father_namefrst",
            "transform": "attach_family_col",
            "other_col": "namefrst_clean",
            "person_pointer": "poploc",
            "family_id": "serial",
            "person_id": "pernum",
        },
        {
            "output_col": "mother_namefrst",
            "transform": "attach_family_col",
            "other_col": "namefrst_clean",
            "person_pointer": "momloc",
            "family_id": "serial",
            "person_id": "pernum",
        },
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)
    pdf_a = (
        spark.table("prepped_df_a")
        .toPandas()[
            [
                "serial",
                "pernum",
                "namefrst_clean",
                "namelast_clean",
                "spouse_namefrst",
                "spouse_bpl",
                "father_namefrst",
                "mother_namefrst",
            ]
        ]
        .sort_values(["serial", "pernum"])
    )
    pdf_b = (
        spark.table("prepped_df_b")
        .toPandas()[
            [
                "serial",
                "pernum",
                "namefrst_clean",
                "namelast_clean",
                "spouse_namefrst",
                "spouse_bpl",
                "father_namefrst",
                "mother_namefrst",
            ]
        ]
        .sort_values(["serial", "pernum"])
    )

    assert (
        pdf_a.query("namefrst_clean == 'jezebel'")["spouse_namefrst"].iloc[0] == "job"
    )
    assert (
        pdf_a.query("namefrst_clean == 'jezebel'")["mother_namefrst"].iloc[0] == "eliza"
    )
    assert (
        pdf_a.query("namefrst_clean == 'willie may'")["father_namefrst"].iloc[0]
        == "wm h"
    )
    assert (
        pdf_a.query("namefrst_clean == 'willie may'")["mother_namefrst"].iloc[0]
        == "martha"
    )

    assert (
        pdf_b.query("namefrst_clean == 'jezebel'")["spouse_namefrst"].iloc[0] == "job"
    )
    assert pdf_b.query("namefrst_clean == 'jezebel'")["spouse_bpl"].iloc[0] == 10
    assert (
        pdf_b.query("namefrst_clean == 'jezebel'")["mother_namefrst"].iloc[0] == "eliza"
    )
    assert pd.isnull(
        pdf_b.query("namefrst_clean == 'esther'")["spouse_namefrst"].iloc[0]
    )


def test_step_1_transform_calc_nativity(
    preprocessing, spark, preprocessing_conf_19thc_nativity_conf
):
    """Test attach_family_col transform on data containing households"""

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas().sort_values(["serial", "pernum"])

    pdf_b = spark.table("prepped_df_b").toPandas().sort_values(["serial", "pernum"])

    assert set(pdf_a["nativity"].tolist()) == {1, 2, 3, 4, 5}
    assert set(pdf_a["test_nativity"].tolist()) == {0}
    assert set(pdf_a["mbpl"].tolist()) == {999}
    assert set(pdf_a["key_mbpl_range_b"].tolist()) == {0}
    assert set(pdf_a["mbpl_range_b"].tolist()) == {0}
    assert pdf_a["mother_nativity"].equals(pdf_a["key_mother_nativity"])
    assert pdf_a["key_mbpl"].equals(pdf_a["mbpl_calc"])
    assert pdf_a["key_fbpl"].equals(pdf_a["fbpl_calc"])
    assert pdf_a["mbpl_range"].equals(pdf_a["key_mbpl_range"])
    assert set(pdf_a["key_mbpl_range"].tolist()) == {0, 1, 2}
    assert pdf_a["key_nativity_calc"].equals(pdf_a["nativity_calc"])

    assert set(pdf_b["nativity"].tolist()) == {0}
    assert set(pdf_b["test_nativity"].tolist()) == {0, 1, 2, 3, 4, 5}
    assert pdf_b["key_nativity_calc"].equals(pdf_b["test_nativity"])
    assert set(pdf_b["key_mbpl_range"].tolist()) == {0}
    assert set(pdf_b["key_mother_nativity"].tolist()) == {0}
    assert pdf_b["mbpl_range_b"].equals(pdf_b["key_mbpl_range_b"])
    assert set(pdf_b["mother_nativity"].tolist()) == {0}
    assert pdf_b["key_mbpl"].equals(pdf_b["mbpl_calc"])
    assert pdf_b["key_fbpl"].equals(pdf_b["fbpl_calc"])
    assert pdf_b["mbpl_range"].equals(pdf_b["mbpl_range_b"])
    assert pdf_b["mbpl_range"].equals(pdf_b["key_mbpl_range_b"])
    assert pdf_b["key_nativity_calc"].equals(pdf_b["nativity_calc"])

    assert pdf_a["nativity_calc"].tolist() == [
        0,
        0,
        1,
        0,
        5,
        0,
        2,
        0,
        5,
        3,
        5,
        0,
        2,
        5,
        5,
        0,
        0,
    ]
    assert pdf_b["nativity_calc"].tolist() == [
        0,
        0,
        1,
        0,
        5,
        0,
        2,
        0,
        5,
        3,
        5,
        5,
        4,
        5,
        5,
        1,
    ]


def test_step_1_transform_related_individuals(
    preprocessing, spark, preprocessing_conf_household_data
):
    """Test attach related_individuals transform"""
    preprocessing_conf_household_data["feature_selections"] = [
        {
            "output_col": "namefrst_related",
            "input_col": "namefrst_clean",
            "transform": "related_individuals",
            "family_id": "serial",
            "relate_col": "relate",
            "top_code": 10,
            "bottom_code": 3,
        }
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = (
        spark.table("prepped_df_a")
        .toPandas()[
            [
                "serial",
                "pernum",
                "relate",
                "namefrst_clean",
                "namelast_clean",
                "namefrst_related",
            ]
        ]
        .sort_values(["serial", "pernum"])
    )
    pdf_b = (
        spark.table("prepped_df_b")
        .toPandas()[
            [
                "serial",
                "pernum",
                "relate",
                "namefrst_clean",
                "namelast_clean",
                "namefrst_related",
            ]
        ]
        .sort_values(["serial", "pernum"])
    )

    assert pdf_a.query("namefrst_clean == 'otillia'")["namefrst_related"].iloc[0] == []
    assert pdf_a.query("namefrst_clean == 'j clauson'")["namefrst_related"].iloc[0] == [
        "eugene"
    ]

    assert sorted(
        pdf_b.query("namefrst_clean == 'job'")["namefrst_related"].iloc[0]
    ) == ["eliza", "jo", "mary"]


def test_step_1_transform_related_individual_rows(
    preprocessing, spark, preprocessing_conf_household_data
):
    """Test attach related_individuals transform"""
    preprocessing_conf_household_data["feature_selections"] = [
        {
            "output_col": "spouse_namefrst",
            "transform": "attach_family_col",
            "other_col": "namefrst_clean",
            "person_pointer": "sploc",
            "family_id": "serial",
            "person_id": "pernum",
        },
        {
            "output_col": "namefrst_related_rows",
            "input_cols": ["namefrst_clean", "bpl", "sex"],
            "transform": "related_individual_rows",
            "family_id": "serial",
            "filters": [
                {"column": "relate", "max": 10, "min": 3},
                {"column": "pernum", "max": 4, "min": 0, "dataset": "b"},
            ],
        },
        {
            "output_col": "unrelated_rows",
            "input_cols": ["namefrst_clean", "bpl", "sex"],
            "transform": "related_individual_rows",
            "family_id": "serial",
            "filters": [{"column": "relate", "max": 999, "min": 11}],
        },
        {
            "output_column": "namelast_neighbors",
            "input_column": "namelast_clean",
            "transform": "neighbor_aggregate",
            "neighborhood_column": "enumdist",
            "sort_column": "serial",
            "range": 5,
        },
    ]

    preprocessing.run_step(0)

    rda = spark.table("raw_df_a")
    rda = (
        rda.withColumn("enumdist_tmp", rda["enumdist"].cast("bigint"))
        .drop("enumdist")
        .withColumnRenamed("enumdist_tmp", "enumdist")
    )
    rda = (
        rda.withColumn("serial_tmp", rda["serial"].cast("bigint"))
        .drop("serial")
        .withColumnRenamed("serial_tmp", "serial")
    )
    rda = (
        rda.withColumn("pernum_tmp", rda["pernum"].cast("bigint"))
        .drop("pernum")
        .withColumnRenamed("pernum_tmp", "pernum")
    )
    rda = (
        rda.withColumn("bpl_tmp", rda["bpl"].cast("bigint"))
        .drop("bpl")
        .withColumnRenamed("bpl_tmp", "bpl")
    )
    rda = (
        rda.withColumn("sex_tmp", rda["sex"].cast("bigint"))
        .drop("sex")
        .withColumnRenamed("sex_tmp", "sex")
    )
    rda = (
        rda.withColumn("relate_tmp", rda["relate"].cast("bigint"))
        .drop("relate")
        .withColumnRenamed("relate_tmp", "relate")
    )

    rda.write.mode("overwrite").saveAsTable("raw_df_a_tmp")
    spark.sql("drop table raw_df_a")
    spark.sql("alter table raw_df_a_tmp rename to raw_df_a")

    rdb = spark.table("raw_df_b")
    rdb = (
        rdb.withColumn("enumdist_tmp", rdb["enumdist"].cast("bigint"))
        .drop("enumdist")
        .withColumnRenamed("enumdist_tmp", "enumdist")
    )
    rdb = (
        rdb.withColumn("serial_tmp", rdb["serial"].cast("bigint"))
        .drop("serial")
        .withColumnRenamed("serial_tmp", "serial")
    )
    rdb = (
        rdb.withColumn("pernum_tmp", rdb["pernum"].cast("bigint"))
        .drop("pernum")
        .withColumnRenamed("pernum_tmp", "pernum")
    )
    rdb = (
        rdb.withColumn("bpl_tmp", rdb["bpl"].cast("bigint"))
        .drop("bpl")
        .withColumnRenamed("bpl_tmp", "bpl")
    )
    rdb = (
        rdb.withColumn("sex_tmp", rdb["sex"].cast("bigint"))
        .drop("sex")
        .withColumnRenamed("sex_tmp", "sex")
    )
    rdb = (
        rdb.withColumn("relate_tmp", rdb["relate"].cast("bigint"))
        .drop("relate")
        .withColumnRenamed("relate_tmp", "relate")
    )

    rdb.write.mode("overwrite").saveAsTable("raw_df_b_tmp")
    spark.sql("drop table raw_df_b")
    spark.sql("alter table raw_df_b_tmp rename to raw_df_b")

    preprocessing.run_step(1)
    select_cols = [
        "serial",
        "pernum",
        "relate",
        "namefrst_clean",
        "namelast_clean",
        "spouse_namefrst",
        "namefrst_related_rows",
        "unrelated_rows",
        "namelast_neighbors",
    ]

    pdf_a = (
        spark.table("prepped_df_a")
        .toPandas()[select_cols]
        .sort_values(["serial", "pernum"])
    )
    pdf_b = (
        spark.table("prepped_df_b")
        .toPandas()[select_cols]
        .sort_values(["serial", "pernum"])
    )

    wilson_a_nbrs = sorted(
        pdf_a.query("namelast_clean == 'wilson'")["namelast_neighbors"].iloc[0]
    )

    assert wilson_a_nbrs == [
        "bierhahn",
        "chambers",
        "cleveland",
        "collins",
        "flemming",
        "graham",
        "harvey",
        "mclean",
        "seward",
        "shields",
    ]

    assert (
        pdf_a.query("namefrst_clean == 'otillia'")["namefrst_related_rows"].iloc[0]
        == []
    )

    row_a = pdf_a.query("namefrst_clean == 'j clauson'")["namefrst_related_rows"].iloc[
        0
    ][0]
    assert row_a.namefrst_clean == "eugene"
    assert row_a.bpl == 10
    assert row_a.sex == 1

    assert (
        pdf_a.query("namefrst_clean == 'jezebel'")["spouse_namefrst"].iloc[0] == "job"
    )

    assert (
        len(pdf_b.query("namefrst_clean == 'job'")["namefrst_related_rows"].iloc[0])
        == 2
    )
    assert (
        pdf_a.query("serial == 2485411 and pernum == 1").unrelated_rows.iloc[0][0][0]
        == "anne"
    )
    assert (
        len(
            pdf_a.query("serial == 2492741 and pernum == 4").namefrst_related_rows.iloc[
                0
            ]
        )
        == 2
    )
    assert (
        len(
            pdf_b.query("serial == 2492741 and pernum == 4").namefrst_related_rows.iloc[
                0
            ]
        )
        == 1
    )


def test_step_1_transform_popularity(
    preprocessing, spark, preprocessing_conf_popularity
):
    """Test attach related_individuals transform"""
    preprocessing_conf_popularity["feature_selections"] = [
        {
            "checkpoint": True,
            "input_cols": ["namefrst", "namelast", "bpl", "sex"],
            "range_col": "birthyr",
            "range_val": 3,
            "output_col": "ncount",
            "transform": "popularity",
        },
        {
            "checkpoint": True,
            "input_cols": ["namefrst", "bpl", "sex", "birthyr"],
            "output_col": "fname_pop",
            "transform": "popularity",
        },
        {
            "checkpoint": True,
            "output_col": "byr_pop",
            "range_col": "birthyr",
            "range_val": 3,
            "transform": "popularity",
        },
        {
            "input_col": "ncount",
            "output_col": "ncount2",
            "transform": "power",
            "exponent": 2,
        },
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").orderBy("id").toPandas()

    assert pdf_a.query("id == 0")["ncount"].iloc[0] == 3
    assert pdf_a.query("id == 1")["ncount"].iloc[0] == 1
    assert pdf_a.query("id == 2")["ncount"].iloc[0] == 2
    assert pdf_a.query("id == 3")["ncount"].iloc[0] == 2
    assert pdf_a.query("id == 4")["ncount"].iloc[0] == 1
    assert pdf_a.query("id == 5")["ncount"].iloc[0] == 1
    assert pdf_a.query("id == 6")["ncount"].iloc[0] == 1
    assert pdf_a.query("id == 7")["ncount"].iloc[0] == 1

    assert pdf_a.query("id == 0")["ncount2"].iloc[0] == 9
    assert pdf_a.query("id == 1")["ncount2"].iloc[0] == 1
    assert pdf_a.query("id == 2")["ncount2"].iloc[0] == 4

    assert pdf_a.query("id == 0")["fname_pop"].iloc[0] == 2
    assert pdf_a.query("id == 1")["fname_pop"].iloc[0] == 1
    assert pdf_a.query("id == 2")["fname_pop"].iloc[0] == 1
    assert pdf_a.query("id == 3")["fname_pop"].iloc[0] == 1
    assert pdf_a.query("id == 4")["fname_pop"].iloc[0] == 2
    assert pdf_a.query("id == 5")["fname_pop"].iloc[0] == 1
    assert pdf_a.query("id == 6")["fname_pop"].iloc[0] == 1
    assert pdf_a.query("id == 7")["fname_pop"].iloc[0] == 1

    assert pdf_a.query("id == 0")["byr_pop"].iloc[0] == 7
    assert pdf_a.query("id == 1")["byr_pop"].iloc[0] == 1
    assert pdf_a.query("id == 2")["byr_pop"].iloc[0] == 4
    assert pdf_a.query("id == 3")["byr_pop"].iloc[0] == 6


def test_step_1_transforms_adv_calc(preprocessing, spark, preprocessing_conf):
    """Test more column mapping with transforms"""
    preprocessing_conf["column_mappings"] = [
        {
            "column_name": "serialp",
            "alias": "serialp_add_a",
            "transforms": [{"type": "concat_to_a", "value": "_a"}],
        },
        {
            "column_name": "bpl",
            "alias": "bpl_add_b",
            "transforms": [{"type": "concat_to_b", "value": 1}],
        },
        {
            "alias": "bpl",
            "column_name": "bpl",
        },
        {
            "alias": "sex",
            "column_name": "sex",
        },
        {
            "alias": "concat_bpl_sex",
            "column_name": "bpl",
            "transforms": [{"type": "concat_two_cols", "column_to_append": "sex"}],
        },
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert pdf_a.query("id == 10")["serialp_add_a"].iloc[0] == "A_a"
    assert pdf_b.query("id == 10")["serialp_add_a"].iloc[0] == "C"
    assert pdf_a.query("id == 20")["bpl_add_b"].iloc[0] == 200
    assert pdf_b.query("id == 30")["bpl_add_b"].iloc[0] == "5001"
    assert pdf_a.query("id == 10")["concat_bpl_sex"].iloc[0] == "1001"
    assert pdf_b.query("id == 50")["concat_bpl_sex"].iloc[0] == "7002"


def test_step_1_transforms_expand(preprocessing, spark, preprocessing_conf):
    """Test transform expand"""
    preprocessing_conf["column_mappings"] = [
        {"column_name": "age"},
        {
            "column_name": "age",
            "alias": "age_expand_3",
            "transforms": [{"type": "expand", "value": 3}],
        },
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()

    assert pdf_a.query("id == 10")["age_expand_3"].iloc[0] == [
        20,
        21,
        22,
        23,
        24,
        25,
        26,
    ]


def test_step_1_override(preprocessing, spark, preprocessing_conf):
    """Test column override"""
    preprocessing_conf["column_mappings"] = [
        {
            "column_name": "serialp",
            "override_column_a": "serialp",
            "override_transforms": [{"type": "lowercase_strip"}],
        }
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert pdf_a.query("id == 30")["serialp"].iloc[0] == "b"
    assert pdf_b.query("id == 50")["serialp"].iloc[0] == "E"


def test_step_1_set_values_a_explicitly(preprocessing, spark, preprocessing_conf):
    """Test setting a column value explicitly"""
    preprocessing_conf["column_mappings"] = [
        {"column_name": "serialp", "alias": "serialp", "set_value_column_a": "c"}
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert pdf_a.query("id == 30")["serialp"].iloc[0] == "c"
    assert pdf_b.query("id == 50")["serialp"].iloc[0] == "E"


def test_step_1_set_values_b_explicitly(preprocessing, spark, preprocessing_conf):
    """Test setting b column value explicitly"""
    preprocessing_conf["column_mappings"] = [
        {"column_name": "serialp", "alias": "serialp", "set_value_column_b": "a"}
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert pdf_a.query("id == 30")["serialp"].iloc[0] == "B"
    assert pdf_b.query("id == 50")["serialp"].iloc[0] == "a"


def test_step_1_substitution(
    preprocessing,
    spark,
    preprocessing_conf_simple_names,
    substitutions_womens_names_path,
    substitutions_mens_names_path,
):
    """Test text substitution"""
    preprocessing_conf_simple_names["column_mappings"] = [
        {
            "column_name": "namefrst",
            "alias": "namefrst_std",
            "transforms": [{"type": "lowercase_strip"}],
        },
        {"column_name": "namefrst"},
        {"column_name": "sex"},
    ]

    preprocessing_conf_simple_names["substitution_columns"] = [
        {
            "column_name": "namefrst_std",
            "substitutions": [
                {
                    "join_column": "sex",
                    "join_value": "1",
                    "substitution_file": substitutions_mens_names_path,
                },
                {
                    "join_column": "sex",
                    "join_value": "2",
                    "substitution_file": substitutions_womens_names_path,
                },
            ],
        }
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert pdf_a.query("id == '10ah'")["namefrst_std"].iloc[0] == "cat"
    assert pdf_a.query("id == '20bc'")["namefrst_std"].iloc[0] == "bernard"
    assert pdf_a.query("id == '34hi'")["namefrst_std"].iloc[0] == "catherine"
    assert pdf_a.query("id == '54de'")["namefrst_std"].iloc[0] == "kat"
    assert pdf_b.query("id == 'c23'")["namefrst_std"].iloc[0] == "bernard"
    assert pdf_b.query("id == 'd45'")["namefrst_std"].iloc[0] == "catherine"
    assert pdf_b.query("id == 'e77'")["namefrst_std"].iloc[0] == "bernard"


def test_step_1_street_abbrev_substitution(
    preprocessing,
    spark,
    preprocessing_conf_street_names,
    substitutions_street_abbrevs_path,
):
    """Test text substitution"""
    preprocessing_conf_street_names["id_column"] = "histid"
    preprocessing_conf_street_names["column_mappings"] = [
        {
            "column_name": "street",
            "alias": "street_unstd",
            "transforms": [
                {"type": "lowercase_strip"},
                {"type": "remove_qmark_hyphen"},
                {"type": "replace_apostrophe"},
                {"type": "condense_strip_whitespace"},
            ],
        },
        {
            "column_name": "street_unstd",
            "alias": "street_swapped",
            "transforms": [
                {
                    "type": "swap_words",
                    "values": {"bch": "beach", "ctr": "center", "rd": "road"},
                }
            ],
        },
    ]

    preprocessing_conf_street_names["substitution_columns"] = [
        {
            "column_name": "street_unstd",
            "substitutions": [
                {
                    "substitution_file": substitutions_street_abbrevs_path,
                    "regex_word_replace": True,
                }
            ],
        }
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()

    assert pdf_a.query("histid == 'a01'")["street_unstd"].iloc[0] == "turnpike 35"
    assert (
        pdf_a.query("histid == 'b02'")["street_unstd"].iloc[0] == "4th terrace avenue"
    )
    assert pdf_a.query("histid == 'c03'")["street_unstd"].iloc[0] == "4th state"
    assert pdf_a.query("histid == 'd04'")["street_unstd"].iloc[0] == "old boulevard"
    assert pdf_a.query("histid == 'e05'")["street_unstd"].iloc[0] == "old motorway"
    assert pdf_a.query("histid == 'f06'")["street_unstd"].iloc[0] == "miami bch road"
    assert pdf_a.query("histid == 'g07'")["street_unstd"].iloc[0] == "center street"
    assert pdf_a.query("histid == 'g08'")["street_unstd"].iloc[0] == "ctr street"
    assert pdf_a.query("histid == 'i09'")["street_unstd"].iloc[0] == "strstreet"

    assert (
        pdf_a.query("histid == 'f06'")["street_swapped"].iloc[0] == "miami beach road"
    )
    assert pdf_a.query("histid == 'g08'")["street_swapped"].iloc[0] == "center street"


def test_step_1_street_remove_stop_words(
    preprocessing,
    spark,
    preprocessing_conf_street_names,
    substitutions_street_abbrevs_path,
):
    """Test text substitution"""
    preprocessing_conf_street_names["id_column"] = "histid"
    preprocessing_conf_street_names["column_mappings"] = [
        {"column_name": "street", "alias": "street_orig"},
        {
            "column_name": "street",
            "alias": "street_unstd",
            "transforms": [
                {"type": "lowercase_strip"},
                {"type": "remove_qmark_hyphen"},
                {"type": "replace_apostrophe"},
                {"type": "condense_strip_whitespace"},
            ],
        },
        {
            "column_name": "street_unstd",
            "alias": "street_removed",
            "transforms": [
                {
                    "type": "remove_stop_words",
                    "values": [
                        "avn",
                        "blvd",
                        "rd",
                        "road",
                        "street",
                        "str",
                        "ter",
                        "trnpk",
                    ],
                },
                {"type": "condense_strip_whitespace"},
            ],
        },
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()

    assert pdf_a.query("histid == 'a01'")["street_removed"].iloc[0] == "35"
    assert pdf_a.query("histid == 'b02'")["street_removed"].iloc[0] == "4th"
    assert pdf_a.query("histid == 'c03'")["street_removed"].iloc[0] == "4th state"
    assert pdf_a.query("histid == 'd04'")["street_removed"].iloc[0] == "old"
    assert pdf_a.query("histid == 'e05'")["street_removed"].iloc[0] == "old motorway"
    assert pdf_a.query("histid == 'f06'")["street_removed"].iloc[0] == "miami bch"
    assert pdf_a.query("histid == 'g07'")["street_removed"].iloc[0] == "centre"
    assert pdf_a.query("histid == 'g08'")["street_removed"].iloc[0] == "ctr"
    assert pdf_a.query("histid == 'i09'")["street_removed"].iloc[0] == "strstreet"


def test_step_1_divide_by_int_mapping_birthyr(
    preprocessing, spark, preprocessing_conf_birthyr
):
    """Test text substitution"""
    preprocessing_conf_birthyr["id_column"] = "histid"
    preprocessing_conf_birthyr["column_mappings"] = [
        {"column_name": "yearp", "alias": "year"},
        {"column_name": "age"},
        {"column_name": "birthyr", "alias": "raw_birthyr"},
        {
            "column_name": "birthyr",
            "transforms": [{"type": "mapping", "mappings": {1999: ""}}],
        },
        {"column_name": "bpl", "alias": "raw_bpl"},
        {
            "column_name": "bpl",
            "transforms": [
                {"type": "divide_by_int", "value": 100},
                {"type": "get_floor"},
            ],
        },
    ]
    preprocessing_conf_birthyr["feature_selections"] = [
        {
            "input_column": "birthyr",
            "output_column": "birthyr_filled",
            "condition": "case when birthyr is null or birthyr == '' then year - age else birthyr end",
            "transform": "sql_condition",
        }
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()

    assert pdf_a.query("histid == 'a01'")["raw_birthyr"].iloc[0] == 1999
    assert pdf_a.query("histid == 'a01'")["birthyr"].iloc[0] == ""
    assert pdf_a.query("histid == 'a01'")["bpl"].iloc[0] == 33
    assert pdf_a.query("histid == 'b00'")["raw_bpl"].iloc[0] == 4799
    assert pdf_a.query("histid == 'b00'")["bpl"].iloc[0] == 47
    assert pdf_a.query("histid == 'b02'")["bpl"].iloc[0] == 30
    assert pdf_a.query("histid == 'd04'")["bpl"].iloc[0] == 1
    assert pdf_a.query("histid == 'a01'")["birthyr_filled"].iloc[0] == "1864"
    assert pdf_a.query("histid == 'b02'")["birthyr_filled"].iloc[0] == "1858"
    assert pd.isnull(pdf_a.query("histid == 'b00'")["birthyr_filled"].iloc[0])
    assert pdf_a.query("histid == 'c03'")["birthyr_filled"].iloc[0] == "1901"
    assert pdf_a.query("histid == 'd04'")["birthyr_filled"].iloc[0] == "1850"


def test_step_1_fix_bpl(preprocessing, spark, preprocessing_conf_birthyr):
    """Test text substitution"""
    preprocessing_conf_birthyr["id_column"] = "histid"
    preprocessing_conf_birthyr["column_mappings"] = [
        {"column_name": "state1"},
        {"column_name": "state2"},
        {"column_name": "bpl", "alias": "bpl_orig"},
        {
            "column_name": "bpl",
            "alias": "bpl_state_orig",
            "transforms": [
                {"type": "divide_by_int", "value": 100},
                {"type": "get_floor"},
            ],
        },
    ]
    preprocessing_conf_birthyr["feature_selections"] = [
        {
            "input_column": "bpl_orig",
            "output_column": "clean_bpl",
            "condition": """case
                            when state1 == "washington" and state2=="washington"
                            then 5300
                            when (state1 is null or state1 == '') and state2=="washington"
                            then 5300
                            when state1 == "washington" and (state2=='' or state2 is null)
                            then 5300
                            else bpl_orig
                            end""",
            "transform": "sql_condition",
        },
        {
            "input_column": "bpl_state_orig",
            "output_column": "bpl_state",
            "condition": """case
                        when state1 == "washington" and state2=="washington"
                        then 53
                        when (state1 is null or state1 == '') and state2=="washington"
                        then 53
                        when state1 == "washington" and (state2=='' or state2 is null)
                        then 53
                        else bpl_state_orig
                        end""",
            "transform": "sql_condition",
        },
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()

    assert pdf_a.query("histid == 'a01'")["clean_bpl"].iloc[0] == 5300
    assert pdf_a.query("histid == 'b02'")["clean_bpl"].iloc[0] == 5300
    assert pdf_a.query("histid == 'b00'")["clean_bpl"].iloc[0] == 4799
    assert pdf_a.query("histid == 'c03'")["clean_bpl"].iloc[0] == 5300
    assert pdf_a.query("histid == 'd04'")["clean_bpl"].iloc[0] == 100
    assert pdf_a.query("histid == 'a01'")["bpl_state"].iloc[0] == 53
    assert pdf_a.query("histid == 'b02'")["bpl_state"].iloc[0] == 53
    assert pdf_a.query("histid == 'b00'")["bpl_state"].iloc[0] == 47
    assert pdf_a.query("histid == 'c03'")["bpl_state"].iloc[0] == 53
    assert pdf_a.query("histid == 'd04'")["bpl_state"].iloc[0] == 1


def test_step_1_bigrams(preprocessing, spark, preprocessing_conf_simple_names):
    """Test checkpoint transform"""
    preprocessing_conf_simple_names["column_mappings"] = [
        {"column_name": "namefrst"},
        {
            "column_name": "namefrst",
            "alias": "namefrst_std",
            "transforms": [{"type": "lowercase_strip"}],
        },
    ]

    preprocessing_conf_simple_names["feature_selections"] = [
        {
            "input_column": "namefrst_std",
            "output_column": "bigrams_namefrst",
            "transform": "bigrams",
        }
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert pdf_a.query("id == '10ah'")["bigrams_namefrst"].iloc[0] == [
        "  c",
        "a t",
        "c a",
    ]
    assert pdf_b.query("id == 'd45'")["bigrams_namefrst"].iloc[0] == [
        "  k",
        "a t",
        "i e",
        "k a",
        "t i",
    ]


def test_step_1_bigrams_no_space(preprocessing, spark, preprocessing_conf_simple_names):
    """Test checkpoint transform"""
    preprocessing_conf_simple_names["column_mappings"] = [
        {"column_name": "namefrst"},
        {
            "column_name": "namefrst",
            "alias": "namefrst_std",
            "transforms": [{"type": "lowercase_strip"}],
        },
    ]

    preprocessing_conf_simple_names["feature_selections"] = [
        {
            "input_column": "namefrst_std",
            "output_column": "bigrams_namefrst",
            "no_first_pad": True,
            "transform": "bigrams",
        }
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert pdf_a.query("id == '10ah'")["bigrams_namefrst"].iloc[0] == ["a t", "c a"]
    assert pdf_b.query("id == 'd45'")["bigrams_namefrst"].iloc[0] == [
        "a t",
        "i e",
        "k a",
        "t i",
    ]


def test_step_1_array(preprocessing, spark, preprocessing_conf_simple_names):
    """Test array transform"""
    preprocessing_conf_simple_names["column_mappings"] = [
        {"column_name": "namefrst"},
        {"column_name": "sex"},
    ]

    preprocessing_conf_simple_names["feature_selections"] = [
        {
            "input_columns": ["namefrst", "sex"],
            "output_column": "namefrst_sex_array",
            "transform": "array",
        }
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert pdf_a.query("id == '54de'")["namefrst_sex_array"].iloc[0] == ["Kat", "1"]
    assert pdf_b.query("id == 'e77'")["namefrst_sex_array"].iloc[0] == ["Bernard", "1"]


def test_step_1_union(
    preprocessing, spark, preprocessing_conf_synthetic_household_data
):
    """Test union transform"""
    preprocessing_conf_synthetic_household_data["column_mappings"] = [
        {"column_name": "namefrst"},
        {"column_name": "namelast"},
        {"column_name": "neighbors"},
        {"column_name": "nonfamily_household"},
    ]

    preprocessing_conf_synthetic_household_data["feature_selections"] = [
        {
            "input_columns": ["neighbors", "nonfamily_household"],
            "output_column": "names_union",
            "transform": "union",
        }
    ]

    preprocessing.run_step(0)
    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert sorted(pdf_a.query("namefrst == 'jane'")["names_union"].iloc[0]) == [
        "edie",
        "elmer",
        "gerald",
    ]
    assert sorted(pdf_a.query("namefrst == 'janice'")["names_union"].iloc[0]) == [
        "edie"
    ]
    assert sorted(pdf_b.query("namefrst == 'gary'")["names_union"].iloc[0]) == [
        "colleen"
    ]


def test_rel_rows_real_data(spark, preprocessing, preprocessing_conf_rel_rows):
    """Test related_rows data and double threshold comparison blocking criteria"""

    preprocessing.run_step(0)

    rda = spark.table("raw_df_a")
    rda = (
        rda.withColumn("serialp_tmp", rda["serialp"].cast("bigint"))
        .drop("serialp")
        .withColumnRenamed("serialp_tmp", "serialp")
    )
    rda = (
        rda.withColumn("relate_tmp", rda["relate"].cast("bigint"))
        .drop("relate")
        .withColumnRenamed("relate_tmp", "relate")
    )
    rda = (
        rda.withColumn("sex_tmp", rda["sex"].cast("bigint"))
        .drop("sex")
        .withColumnRenamed("sex_tmp", "sex")
    )
    rda = (
        rda.withColumn("age_tmp", rda["age"].cast("bigint"))
        .drop("age")
        .withColumnRenamed("age_tmp", "age")
    )

    rda.write.mode("overwrite").saveAsTable("raw_df_a_tmp")
    spark.sql("drop table raw_df_a")
    spark.sql("alter table raw_df_a_tmp rename to raw_df_a")

    rdb = spark.table("raw_df_b")
    rdb = (
        rdb.withColumn("serialp_tmp", rdb["serialp"].cast("bigint"))
        .drop("serialp")
        .withColumnRenamed("serialp_tmp", "serialp")
    )
    rdb = (
        rdb.withColumn("relate_tmp", rdb["relate"].cast("bigint"))
        .drop("relate")
        .withColumnRenamed("relate_tmp", "relate")
    )
    rdb = (
        rdb.withColumn("sex_tmp", rdb["sex"].cast("bigint"))
        .drop("sex")
        .withColumnRenamed("sex_tmp", "sex")
    )
    rdb = (
        rdb.withColumn("age_tmp", rdb["age"].cast("bigint"))
        .drop("age")
        .withColumnRenamed("age_tmp", "age")
    )

    rdb.write.mode("overwrite").saveAsTable("raw_df_b_tmp")
    spark.sql("drop table raw_df_b")
    spark.sql("alter table raw_df_b_tmp rename to raw_df_b")

    preprocessing.run_step(1)

    pdf_a = spark.table("prepped_df_a").toPandas()
    pdf_b = spark.table("prepped_df_b").toPandas()

    assert (
        len(
            pdf_a.query("histid == 'D1DAEB8F-66F0-435C-8E45-F004D967549D'")[
                "namefrst_related_rows"
            ].iloc[0]
        )
        == 3
    )
    assert (
        len(
            pdf_a.query("histid == 'D1DAEB8F-66F0-435C-8E45-F004D967549D'")[
                "namefrst_related_rows_age_min_5"
            ].iloc[0]
        )
        == 1
    )
    assert (
        len(
            pdf_a.query("histid == 'D1DAEB8F-66F0-435C-8E45-F004D967549D'")[
                "namefrst_related_rows_age_b_min_5"
            ].iloc[0]
        )
        == 3
    )

    assert (
        len(
            pdf_b.query("histid == 'B04F6A33-9A86-4EAF-884B-0BD6107CCDEB'")[
                "namefrst_related_rows"
            ].iloc[0]
        )
        == 7
    )
    assert (
        len(
            pdf_b.query("histid == 'B04F6A33-9A86-4EAF-884B-0BD6107CCDEB'")[
                "namefrst_related_rows_age_min_5"
            ].iloc[0]
        )
        == 6
    )
    assert (
        len(
            pdf_b.query("histid == 'B04F6A33-9A86-4EAF-884B-0BD6107CCDEB'")[
                "namefrst_related_rows_age_b_min_5"
            ].iloc[0]
        )
        == 6
    )

    assert (
        len(
            pdf_a.query("histid == '8B0A8FA5-A260-4841-95D0-2C45689485C8'")[
                "namefrst_related_rows"
            ].iloc[0]
        )
        == 5
    )
    assert (
        len(
            pdf_a.query("histid == '8B0A8FA5-A260-4841-95D0-2C45689485C8'")[
                "namefrst_related_rows_age_min_5"
            ].iloc[0]
        )
        == 2
    )
    assert (
        len(
            pdf_a.query("histid == '8B0A8FA5-A260-4841-95D0-2C45689485C8'")[
                "namefrst_related_rows_age_b_min_5"
            ].iloc[0]
        )
        == 5
    )

    assert (
        len(
            pdf_b.query("histid == 'F7E0450D-ECCC-4338-92B0-ACB4F9D40D8F'")[
                "namefrst_related_rows"
            ].iloc[0]
        )
        == 7
    )
    assert (
        len(
            pdf_b.query("histid == 'F7E0450D-ECCC-4338-92B0-ACB4F9D40D8F'")[
                "namefrst_related_rows_age_min_5"
            ].iloc[0]
        )
        == 7
    )
    assert (
        len(
            pdf_b.query("histid == 'F7E0450D-ECCC-4338-92B0-ACB4F9D40D8F'")[
                "namefrst_related_rows_age_b_min_5"
            ].iloc[0]
        )
        == 7
    )



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import subprocess
import pyspark.sql.functions as pyspark_funcs


def export_crosswalk(spark, output_path, variables, include_round):
    crosswalk_vars = ["histid_a", "histid_b"]

    if include_round:
        crosswalk_vars.append("round")

    if "histid" not in variables:
        variables.append("histid")

    # We are accessing these records in order to attach extra contextual
    # variables to the pairs of people in the predicted matches tables.
    raw_df_a = spark.table("raw_df_a")
    raw_df_b = spark.table("raw_df_b")

    # We are adding (unioning) together individual matches (predicted_matches) and
    # household matches (hh_predicted_matches)
    #
    # The addition of 'round' is to identify which matches came from which matching round.
    # We may or may not select and export it depending on the --include-round flag.
    predicted_matches = spark.table("predicted_matches").withColumn(
        "round", pyspark_funcs.lit(1)
    )

    hh_predicted_matches = spark.table("hh_predicted_matches").withColumn(
        "round", pyspark_funcs.lit(2)
    )

    for variable in variables:
        if variable not in [c.lower() for c in raw_df_a.columns]:
            print(f"Error: variable '{variable}' does not exist in raw_df_a.")
            return
        if variable not in [c.lower() for c in raw_df_b.columns]:
            print(f"Error: variable '{variable}' does not exist in raw_df_a.")
            return

    all_matches = predicted_matches.select(crosswalk_vars).unionByName(
        hh_predicted_matches.select(crosswalk_vars)
    )

    # Make distinct sets of variable names for the a and b datasets
    columns_a = [f"{variable} as {variable}_a" for variable in variables]
    columns_b = [f"{variable} as {variable}_b" for variable in variables]

    raw_df_a_selected = raw_df_a.selectExpr(columns_a)
    raw_df_b_selected = raw_df_b.selectExpr(columns_b)

    all_matches_with_selections = all_matches.join(raw_df_a_selected, "histid_a").join(
        raw_df_b_selected, "histid_b"
    )

    if "csv" in output_path.split("."):
        export_csv(all_matches_with_selections, output_path)
    else:
        export_fixed_width(variables, all_matches_with_selections, output_path)


def export_csv(all_matches_with_selections, output_path):
    output_tmp = output_path + ".tmp"
    all_matches_with_selections.write.csv(output_tmp, header=False)
    header = (
        '"' + '","'.join([col.name for col in all_matches_with_selections.schema]) + '"'
    )

    commands = [
        f"echo '{header}' > {output_path}",
        f"cat {output_tmp}/* >> {output_path} ",
        f"rm -rf {output_tmp}",
    ]

    for command in commands:
        subprocess.run(command, shell=True)


def export_fixed_width(variables, all_matches_with_selections, output_path):
    output_tmp = output_path + ".tmp"
    sizes = {
        "histid": 36,
        "serialp": 8,
        "pernum": 4,
        "age": 3,
        "sex": 1,
        "statefip_p": 2,
        "bpl": 5,
    }
    fw_columns_a = []
    fw_columns_b = []
    for variable in variables:
        size = sizes.get(variable, 15)
        fw_columns_a.append(
            [f"LPAD({variable}_a, {size}, ' ') as {variable}_a", size, f"{variable}_a"]
        )
    fw_columns_b.append(
        [f"LPAD({variable}_b, {size}, ' ') as {variable}_b", size, f"{variable}_b"]
    )
    all_column_selects = [c[0] for c in (fw_columns_a + fw_columns_b)]

    [print(f"{c[2]} - {c[1]}") for c in (fw_columns_a + fw_columns_b)]
    all_matches_fixed_width = all_matches_with_selections.selectExpr(all_column_selects)
    all_matches_fixed_width.selectExpr("CONCAT_WS('', *)").write.text(output_tmp)

    commands = [f"cat {output_tmp}/* >> {output_path} ", f"rm -rf {output_tmp}"]
    for command in commands:
        subprocess.run(command, shell=True)



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from hlink.linking.link_task import LinkTask


def tfam(link_run, id_col, id_a, id_b):
    cols = [
        f"{id_col}",
        "serialp",
        "pernum",
        "relate",
        "namefrst",
        "namelast",
        "age",
        "birthyr",
        "sex",
        "race",
        "marst",
        "durmarr",
        "bpl",
        "nativity",
        "citizen",
        "mbpl",
        "fbpl",
        "statefip_p",
        "street_p",
        "countyicp_p",
        "region_p",
    ]

    if (
        link_run.get_table("training_data").exists()
        and link_run.get_table("training_features").exists()
    ):
        pass
    else:
        link_run.training.run_step(0)
        link_run.training.run_step(1)

    table_a, table_b = _prep_tfam_tables(
        link_run, "tfam_table", "training_data", cols, id_col
    )

    serialp_a = table_a.filter(f"{id_col} == '{id_a}'").take(1)[0]["serialp"]
    serialp_b = table_b.filter(f"{id_col} == '{id_b}'").take(1)[0]["serialp"]
    print("Family 1900:")
    table_a.where(f"serialp = '{serialp_a}'").orderBy("pernum").select(cols).show(
        truncate=False
    )
    print("Family 1910:")
    table_b.where(f"serialp = '{serialp_b}'").orderBy("pernum").select(cols).show(
        truncate=False
    )

    true_match = (
        link_run.spark.table("training_data")
        .where(f"histid_a = '{id_a}' AND histid_b != '{id_b}' AND match = 1")
        .take(1)
    )

    if len(true_match) != 0:
        print("Match family 1910:")
        histid_b_tm = true_match[0][f"{id_col}_b"]
        print(histid_b_tm)
        serialp_matched = table_b.filter(f"{id_col} == '{histid_b_tm}'").take(1)[0][
            "serialp"
        ]
        table_b.where(f"serialp = '{serialp_matched}'").orderBy("pernum").select(
            cols
        ).show(truncate=False)

    print("All hits:")

    tfam_table = link_run.get_table("tfam_hits")
    if tfam_table.exists():
        table_hits = tfam_table.df()
    else:
        sql = """
                    select pa.histid as histid_a, pb.histid as histid_b, pa.namefrst as namefrst_a, pa.namelast as namelast_a, pb.namefrst as namefrst_b, pb.namelast as namelast_b, pa.mbpl as mbpl_a, pb.mbpl as mbpl_b, pa.fbpl as fbpl_a, pb.fbpl as fbpl_b, pa.statefip_p as state_a, pb.statefip_p as state_b, pa.countyicp_p as county_a, pb.countyicp_p as county_b, tf.namefrst_jw, tf.namelast_jw
                    from training_features tf
                    left join raw_df_a pa on pa.histid = tf.histid_a
                    join raw_df_b pb on pb.histid = tf.histid_b
                    """
        table_hits = LinkTask(link_run).run_register_sql(
            name=tfam_table.name, sql=sql, persist=True
        )

    table_hits.where(f"{id_col}_a == '{id_a}'").orderBy(
        ["namelast_jw", "namefrst_jw"], ascending=False
    ).show(20, False)

    print("Features:")
    link_run.spark.table("training_features").where(
        f"histid_a = '{id_a}' AND histid_b = '{id_b}'"
    ).show(100, False)


def tfam_raw(link_run, id_col, id_a, id_b):
    cols = [
        f"{id_col}",
        "serial_p",
        "pernum",
        "relate",
        "namefrst",
        "namelast",
        "age",
        "birthyr",
        "sex",
        "race",
        "marst",
        "durmarr",
        "bpl",
        "nativity",
        "citizen",
        "mbpl",
        "fbpl",
        "statefip_p",
        "street_p",
        "countyicp_p",
        "region_p",
    ]

    table_a = link_run.spark.table("raw_df_a")
    table_b = link_run.spark.table("raw_df_b")

    serialp_a = table_a.filter(f"{id_col} == '{id_a}'").take(1)[0]["SERIAL_P"]
    serialp_b = table_b.filter(f"{id_col} == '{id_b}'").take(1)[0]["SERIAL_P"]
    print("Family 1900:")
    table_a.where(f"SERIAL_P = '{serialp_a}'").orderBy("PERNUM").select(cols).show(
        truncate=False
    )
    print("Family 1910:")
    table_b.where(f"SERIAL_P = '{serialp_b}'").orderBy("PERNUM").select(cols).show(
        truncate=False
    )


def hh_tfam(link_run, id_col, id_a, id_b):
    cols = [
        f"{id_col}",
        "serialp",
        "pernum",
        "relate",
        "namefrst",
        "namelast",
        "age",
        "birthyr",
        "sex",
        "race",
        "marst",
        "durmarr",
        "bpl",
        "nativity",
        "citizen",
        "mbpl",
        "fbpl",
        "statefip_p",
        "street_p",
        "countyicp_p",
        "region_p",
    ]

    table_a, table_b = _prep_tfam_tables(
        link_run, "hh_tfam_table", "hh_predicted_matches", cols, id_col
    )

    serialp_a = table_a.filter(f"{id_col} == '{id_a}'").take(1)[0]["serialp"]
    serialp_b = table_b.filter(f"{id_col} == '{id_b}'").take(1)[0]["serialp"]
    print("Family 1900:")
    table_a.where(f"serialp = '{serialp_a}'").orderBy("pernum").select(cols).show(
        truncate=False
    )
    print("Family 1910:")
    table_b.where(f"serialp = '{serialp_b}'").orderBy("pernum").select(cols).show(
        truncate=False
    )


def hh_tfam_2a(link_run, id_col, id_a1, id_a2, id_b):
    cols = [
        "serialp",
        "pernum",
        "relate",
        "age",
        "birthyr",
        "sex",
        "race",
        "marst",
        "durmarr",
        "bpl",
        "nativity",
        "citizen",
        "birthyr",
        "namelast",
        "namefrst",
        "mbpl",
        "fbpl",
        "statefip_p",
        "street_p",
        "countyicp_p",
        "region_p",
        "histid",
    ]

    table_a, table_b = _prep_tfam_tables(
        link_run, "hh_tfam_table", "hh_predicted_matches", cols, id_col
    )

    serialp_a1 = table_a.filter(f"{id_col} == '{id_a1}'").take(1)[0]["serialp"]
    serialp_a2 = table_a.filter(f"{id_col} == '{id_a2}'").take(1)[0]["serialp"]
    serialp_b = table_b.filter(f"{id_col} == '{id_b}'").take(1)[0]["serialp"]
    print("Family 1900 option 1:")
    table_a.where(f"serialp = '{serialp_a1}'").orderBy("pernum").select(cols).show(
        truncate=False
    )
    print("Family 1900 option 2:")
    table_b.where(f"serialp = '{serialp_a2}'").orderBy("pernum").select(cols).show(
        truncate=False
    )
    print("Family 1910:")
    table_b.where(f"serialp = '{serialp_b}'").orderBy("pernum").select(cols).show(
        truncate=False
    )


def hh_tfam_2b(link_run, id_col, id_a, id_b1, id_b2):
    cols = [
        f"{id_col}",
        "serialp",
        "pernum",
        "relate",
        "namefrst",
        "namelast",
        "age",
        "birthyr",
        "sex",
        "race",
        "marst",
        "durmarr",
        "bpl",
        "nativity",
        "citizen",
        "mbpl",
        "fbpl",
        "statefip_p",
        "street_p",
        "countyicp_p",
        "region_p",
    ]

    table_a, table_b = _prep_tfam_tables(
        link_run, "hh_tfam_table", "hh_predicted_matches", cols, id_col
    )

    serialp_a = table_a.filter(f"{id_col} == '{id_a}'").take(1)[0]["serialp"]
    serialp_b1 = table_b.filter(f"{id_col} == '{id_b1}'").take(1)[0]["serialp"]
    serialp_b2 = table_b.filter(f"{id_col} == '{id_b2}'").take(1)[0]["serialp"]
    print("Family 1900:")
    table_a.where(f"serialp = '{serialp_a}'").orderBy("pernum").select(cols).show(
        truncate=False
    )
    print("Family 1910 option 1:")
    table_b.where(f"serialp = '{serialp_b1}'").orderBy("pernum").select(cols).show(
        truncate=False
    )
    print("Family 1910 option 2:")
    table_b.where(f"serialp = '{serialp_b2}'").orderBy("pernum").select(cols).show(
        truncate=False
    )


def _prep_tfam_tables(link_run, table_name, source_table, cols, id_col):
    tables = []
    for a_or_b in ["a", "b"]:
        table = link_run.get_table(f"{table_name}_{a_or_b}")
        if table.exists():
            tables.append(table.df())
        else:
            tables.append(
                LinkTask(link_run).run_register_sql(
                    name=table.name,
                    template="tfam_tables",
                    t_ctx={
                        "a_or_b": a_or_b,
                        "cols": cols,
                        "id": id_col,
                        "source_table": f"{source_table}",
                    },
                    persist=True,
                )
            )
    return tables[0], tables[1]
